from typing import List, Union

from vidur.entities import Batch, KVParallelBatch
from vidur.events import BaseEvent
from vidur.logger import init_logger
from vidur.metrics import MetricsStore
from vidur.scheduler import BaseGlobalScheduler
from vidur.types import EventType

logger = init_logger(__name__)


class BatchStageArrivalEvent(BaseEvent):
    def __init__(
        self,
        time: float,
        replica_id: int,
        stage_id: int,
        batch: Union[Batch, KVParallelBatch],
    ):
        super().__init__(time, EventType.BATCH_STAGE_ARRIVAL)

        self._replica_id = replica_id
        self._stage_id = stage_id
        self._batch = batch

    def handle_event(
        self, scheduler: BaseGlobalScheduler, metrics_store: MetricsStore
    ) -> List[BaseEvent]:
        from vidur.events.replica_stage_schedule_event import ReplicaStageScheduleEvent

        scheduler.get_replica_stage_scheduler(
            self._replica_id, self._stage_id
        ).add_batch(self._batch)

        return [
            ReplicaStageScheduleEvent(
                self.time,
                self._replica_id,
                self._stage_id,
            )
        ]

    def to_dict(self):
        return {
            "time": self.time,
            "event_type": self.event_type,
            "replica_id": self._replica_id,
            "stage_id": self._stage_id,
            "batch_id": self._batch.id,
        }



import streamlit as st

from vidur.config_optimizer.analyzer.constants import AXIS_COLS
from vidur.config_optimizer.analyzer.dashboard.utils import add_small_divider, plot_cdf


def render_search_instances(df):
    cols = st.columns([1] * len(df))

    for idx, (_, config_row) in enumerate(df.iterrows()):
        col = cols[idx]
        with col:
            container = st.container(border=True)
            with container:
                config_name = (
                    f"QPS: {config_row['poisson_request_interval_generator_qps']:.2f}"
                )
                st.markdown(f"### {config_name}")

                add_small_divider()

                st.markdown("##### Configuration Details")
                for axis_col, value in config_row[AXIS_COLS.values()].items():
                    st.write(f"**{axis_col}**: {value}")

                add_small_divider()

                st.markdown("##### Performance Metrics")

                # extract config hash - it is part of the output dir
                # .*/runs/<config_hash>/.*
                # first find /runs/ and then take the next part
                config_hash = config_row["output_dir"].split("/runs/")[1].split("/")[0]

                st.write(f"**Config Hash**: {config_hash}")
                st.write(
                    f"**Scheduling Delay (P95)**: {config_row['request_scheduling_delay_95%']:.2f} s"
                )
                st.write(f"**TTFT (P95)**: {config_row['TTFT_95%']:.2f} s")
                st.write(f"**TBT (P99)**: {config_row['TBT_99%']:.2f} ms")
                st.write(f"**QPS per Dollar**: {config_row['QPS per Dollar']:.2f}")
                st.write(f"**MFU**: {config_row['mfu_mean']:.2f}%")
                st.write(f"**Memory Usage**: {config_row['memory_usage_mean']:.2f}%")
                st.write(f"**Busy Time**: {config_row['busy_time_percent_mean']:.2f}%")

    add_small_divider()

    df["QPS"] = df["poisson_request_interval_generator_qps"]

    st.markdown("### CDFs")
    plot_cdf(df, y="TTFT_cdf", y_name="TTFT", color="QPS")
    plot_cdf(df, y="TBT_cdf", y_name="TBT", color="QPS")
    plot_cdf(df, y="batch_size_cdf", y_name="Batch Size", color="QPS")
    plot_cdf(df, y="batch_num_tokens_cdf", y_name="Batch Num Tokens", color="QPS")


def render_search_analysis_page(df):
    st.markdown("## Search Analysis")
    st.markdown("---")
    st.markdown("This tool helps in analysis of search sequence.")

    st.markdown("--- \n\n ### Select configurations to analyze")

    COMPARISON_COLS = ["Model", "Trace"] + list(AXIS_COLS.values())

    with st.form(key="search_analysis_form"):
        col1, col2 = st.columns(2)

        config_filters = {}
        for i, axis_col in enumerate(COMPARISON_COLS):
            col = col1 if i % 2 == 0 else col2
            with col:
                config_filters[axis_col] = st.selectbox(
                    axis_col, df[axis_col].unique(), key=f"{axis_col}_selector"
                )

        # apply filters
        filtered_df = df
        for axis_col, value in config_filters.items():
            filtered_df = filtered_df[filtered_df[axis_col] == value]

        # Submit button for the form
        submit_button = st.form_submit_button(label="Add Search Runs")

    if submit_button:
        st.markdown("--- \n\n")
        st.markdown("### Results")

        render_search_instances(filtered_df)



class BaseEntity:
    _id = -1

    @classmethod
    def generate_id(cls):
        cls._id += 1
        return cls._id

    @property
    def id(self) -> int:
        return self._id

    def __str__(self) -> str:
        # use to_dict to get a dict representation of the object
        # and convert it to a string
        class_name = self.__class__.__name__
        return f"{class_name}({str(self.to_dict())})"

    def __repr__(self) -> str:
        return self.__str__()



from queue import PriorityQueue


class EventQueue:
    def __init__(self):
        self.event_queue = PriorityQueue()

    def empty(self):
        return self.event_queue.empty()

    def put(self, event):
        self.event_queue.put(event)

    def get(self):
        return self.event_queue.get()

    def __len__(self):
        return self.event_queue.qsize()



from typing import Callable

import torch

WARMUP_STEPS = 5
GRAPH_STEPS = 3


class GraphedCollective:
    def __init__(
        self,
        num_workers: int,
        size: int,
        collective: str = "all_reduce",
        disable_graph: bool = False,
        dtype: torch.dtype = torch.float16,
    ) -> None:
        self._size = size
        self._disable_graph = disable_graph
        self._collective_fn = self._get_collective_fn(collective)

        self._buffer = torch.empty(
            size=(size,),
            dtype=dtype,
            device="cuda",
        )
        self._gather_buffer = None
        if collective == "all_gather":
            self._gather_tensor = torch.empty(
                size=(size * num_workers,),
                dtype=dtype,
                device="cuda",
            )
        elif collective == "reduce_scatter":
            self._reduce_buffer = torch.empty(
                size=(size * num_workers,),
                dtype=dtype,
                device="cuda",
            )

        if not self._disable_graph:
            self._graph = self._build_graph()

    def _run_all_reduce(self):
        torch.distributed.all_reduce(self._buffer)

    def _run_all_gather(self):
        torch.distributed.all_gather_into_tensor(self._gather_tensor, self._buffer)

    def _run_broadcast(self):
        torch.distributed.broadcast(self._buffer, 0)

    def _run_send_recv(self):
        if torch.distributed.get_rank() == 0:
            torch.distributed.send(self._buffer, 1)
        else:
            torch.distributed.recv(self._buffer, 0)

    def _run_reduce_scatter(self):
        torch.distributed.reduce_scatter_tensor(self._buffer, self._reduce_buffer)

    def _get_collective_fn(self, collective: str) -> Callable:
        if collective == "all_reduce":
            return self._run_all_reduce
        elif collective == "all_gather":
            return self._run_all_gather
        elif collective == "broadcast":
            return self._run_broadcast
        elif collective == "send_recv":
            return self._run_send_recv
        elif collective == "reduce_scatter":
            return self._run_reduce_scatter
        else:
            raise ValueError(f"Unknown collective: {collective}")

    def _build_graph(self) -> torch.cuda.CUDAGraph:
        # Warm up.
        for _ in range(WARMUP_STEPS):
            self._collective_fn()

        torch.cuda.synchronize()

        # Build graph.
        graph = torch.cuda.CUDAGraph()

        mempool = torch.cuda.graph_pool_handle()

        with torch.profiler.profile(
            activities=[torch.profiler.ProfilerActivity.CPU],
        ):
            with torch.cuda.graph(graph, mempool):
                for _ in range(GRAPH_STEPS):
                    self._collective_fn()

        torch.cuda.synchronize()
        return graph

    def launch(self) -> torch.Tensor:
        # NOTE: x must be a slice of self._buffer.
        if self._disable_graph:
            self._collective_fn()
        else:
            self._graph.replay()



from math import ceil
from typing import List

import numpy as np
import sarathi.metrics.cuda_timer
import torch

from vidur.profiling.common.cuda_timer import CudaTimer

# monkey patching the CudaTimer class to use the sarathi implementation
sarathi.metrics.cuda_timer.CudaTimer = CudaTimer

from sarathi.config import CacheConfig, ParallelConfig, SarathiSchedulerConfig
from sarathi.model_executor.attention import (
    AttentionBackend,
    create_attention_wrapper,
    get_attention_wrapper,
    set_attention_backend,
)

from vidur.profiling.attention.attention_input import AttentionInput
from vidur.profiling.attention.sequence_proxy import SequenceMetadataProxy
from vidur.profiling.common.model_config import ModelConfig
from vidur.profiling.common.timer_stats_store import TimerStatsStore

WARMUP_STEPS = 2
ACTIVE_STEPS = 5
MAX_NUM_SEQS = 1


class AttentionWrapper:
    def __init__(
        self,
        model_config: ModelConfig,
        parallel_config: ParallelConfig,
        max_num_blocks: int,
        max_model_len: int,
        block_size: int,
        attention_backend: AttentionBackend,
        dtype: torch.dtype,
    ):
        self.time_stats_store = TimerStatsStore(profile_method="CUDA_EVENT")

        self._model_config = model_config
        self._model_config.max_model_len = max_model_len
        self._parallel_config = parallel_config
        self._dtype = dtype
        self._device = torch.device("cuda")

        self._max_model_len = max_model_len
        self._n_worker_q_heads = self._model_config.get_num_q_heads(
            self._parallel_config
        )
        self._n_worker_kv_heads = self._model_config.get_num_kv_heads(
            self._parallel_config
        )
        self._head_dim = self._model_config.get_head_size()

        self._block_size = block_size

        self._attention_backend = attention_backend
        set_attention_backend(attention_backend)
        scheduler_config = SarathiSchedulerConfig(
            max_num_seqs=MAX_NUM_SEQS,
        )
        create_attention_wrapper(
            scheduler_config,
            self._model_config,
            self._parallel_config,
            self._block_size,
            self._device,
        )
        self._max_blocks_per_sequence = ceil(max_model_len / self._block_size)
        # We create (big) KV tensors and reuse them
        self.max_num_blocks = max_num_blocks
        self.kv_cache = get_attention_wrapper().get_cache_block(
            self.max_num_blocks, dtype=self._dtype, device=self._device
        )
        cache_config = CacheConfig(
            num_gpu_blocks=max_num_blocks,
            block_size=self._block_size,
        )
        get_attention_wrapper().set_cache_config(cache_config)

    def _get_input_tensors(
        self,
        attention_input: AttentionInput,
    ):
        num_tokens_per_seq = (
            attention_input.prefill_chunk_size if attention_input.is_prefill else 1
        )
        batch_size = attention_input.batch_size
        query = torch.randn(
            batch_size * num_tokens_per_seq,
            self._n_worker_q_heads * self._head_dim,
            dtype=self._dtype,
            device=self._device,
        )
        key = torch.randn(
            batch_size * num_tokens_per_seq,
            self._n_worker_kv_heads * self._head_dim,
            dtype=self._dtype,
            device=self._device,
        )
        value = torch.randn(
            batch_size * num_tokens_per_seq,
            self._n_worker_kv_heads * self._head_dim,
            dtype=self._dtype,
            device=self._device,
        )
        # Create SequenceMetadataProxy objects corresponding to AttentionInput
        seq_metadata_list: List[SequenceMetadataProxy] = []
        for _ in range(attention_input.batch_size):
            num_blocks = ceil(
                (num_tokens_per_seq + attention_input.kv_cache_size) / self._block_size
            )
            # TODO(nitinkedia7): Investigate why high=max_num_blocks fails with a CUDA illegal memory access
            seq_metadata = SequenceMetadataProxy(
                is_prompt=attention_input.is_prefill,
                total_len=num_tokens_per_seq + attention_input.kv_cache_size,
                processed_len=attention_input.kv_cache_size,
                block_table=np.arange(num_blocks, dtype=np.int32),
            )
            seq_metadata_list.append(seq_metadata)
        return seq_metadata_list, query, key, value, self.kv_cache

    @torch.inference_mode()
    def profile(
        self,
        attention_input: AttentionInput,
    ):
        # batch size is always 1 for prefill and can be different for decode
        assert attention_input.is_valid(self._max_model_len)

        seq_metadata_list, query, key, value, kv_cache = self._get_input_tensors(
            attention_input,
        )
        get_attention_wrapper().begin_forward(seq_metadata_list)

        for _ in range(WARMUP_STEPS):
            get_attention_wrapper().forward(query, key, value, kv_cache)
        torch.cuda.synchronize()

        self.time_stats_store.clear_stats()

        for _ in range(ACTIVE_STEPS):
            get_attention_wrapper().forward(query, key, value, kv_cache)
        torch.cuda.synchronize()

        get_attention_wrapper().end_forward()

        return {
            "time_stats": self.time_stats_store.get_stats(),
            "n_embd": self._model_config.embedding_dim,
            "n_q_head": self._model_config.num_q_heads,
            "n_kv_head": self._model_config.num_kv_heads,
            "block_size": self._block_size,
            "num_tensor_parallel_workers": self._parallel_config.tensor_parallel_size,
            "max_model_len": self._max_model_len,
            "batch_size": attention_input.batch_size,
            "prefill_chunk_size": attention_input.prefill_chunk_size,
            "kv_cache_size": attention_input.kv_cache_size,
            "is_prefill": attention_input.is_prefill,
            "attention_backend": self._attention_backend,
        }



import streamlit as st

from vidur.config_optimizer.analyzer.dashboard.utils import add_small_divider


def render_intro_page():
    st.markdown("## Vidur Demo: Config Explorer")
    add_small_divider()
    st.markdown(
        "Vidur helps you explore how different system configuration parameters affect LLM inference performance. To know more about Vidur refer to our [MLSys24 paper](https://mlsys.org/virtual/2024/poster/2667) and check out the [source code](https://github.com/microsoft/vidur)."
    )
    add_small_divider()
    st.markdown("### Tools")
    st.markdown(
        "We provide four tools to help you analyze the results of Vidur simulation runs:"
    )
    st.markdown("#### Best Config Selection")
    st.markdown(
        "This tool helps you find and visualize variations in best configuration for a given model model under different workloads."
    )
    st.markdown("#### Config Comparison")
    st.markdown(
        'This tool enabled one-on-one performance comparison between different configurations. You can either manually add configuration on this page or use the **"Add to Configuration"** button on other pages to quickly populate the comparison.'
    )
    st.markdown("#### Pareto Curve Analysis")
    st.markdown(
        "This tool helps you visualize the trade-offs between different performance metrics and helps you identify the best configurations."
    )
    st.markdown("#### Cost Analysis")
    st.markdown(
        "This tool helps you select the best configuration based on your SLOs and cost budget. It also provides additional visualizations which allow to see the performance as a function of different config options."
    )
    add_small_divider()
    st.markdown("### Sample Models and Workloads")
    st.markdown(
        "We provide a set of sample models and workloads to help you get started with Vidur. These models and workloads are based on real-world datasets and are designed to help you understand how different configurations affect performance."
    )
    st.markdown("#### Models")
    st.markdown("We provide the following models for you to explore in Vidur:")
    st.markdown(
        """
        | Model Name | Number of Parameters | Number of Layers | Embedding Size | Number of Attention Heads | Attention Type |
        |------------|-------------|-----------------|---------------|--------------------------|---------------|
        | phi-2 | 2.7B | 32| 2560 | 32 | Multi-Head Attention |
        | Llama-2 7B | 7B | 32 | 4096 | 32 | Multi-Head Attention |
        | InternLM 20B | 20B | 60 | 5120 | 40 | Multi-Head Attention |
        | CodeLlama 34B | 34B | 48 | 8192 | 64 | Group-Head Attention |
        | Llama-2 70B | 70B | 80 | 8192 | 64 | Group-Head Attention |
        | Qwen-72B | 72B | 80 | 8192 | 64 | Multi-Head Attention |
        """
    )
    st.markdown("#### Workloads")
    st.markdown(
        "We provide the following three different workloads for you to explore in Vidur:"
    )
    st.markdown(
        """
        | Dataset                       | Content                                           | Num queries | Num prefill tokens (mean, median, P90) | Num decode tokens (mean, median, P90) | PD Ratio (median, std dev) |
        |-------------------------------|---------------------------------------------------|-----------|--------------------------------------|-------------------------------------|----------------------------|
        | LMSys-Chat-1M-4K      | Natural language conversations | 2M        | 686, 417, 1678                       | 197, 139, 484                       | 2.3, 228                   |
        | Arxiv-Summarization-4K | Summarization of arXiv papers  | 28k       | 2588, 2730, 3702                    | 291, 167, 372                       | 15.7, 16                   |
        | Bilingual-Web-Book-4K  | Document-level English–Chinese translation dataset | 33k       | 1067, 1037, 1453                    | 1612, 1601, 2149                    | 0.65, 0.37                 |
        """
    )
    st.markdown("")
    add_small_divider()
    st.markdown("### Citation")
    st.markdown("If you use Vidur in your research, please cite the following paper:")
    st.markdown(
        """
        ```
        @article{agrawal2024vidur,
          title={Vidur: A Large-Scale Simulation Framework For LLM Inference},
          author={Agrawal, Amey and Kedia, Nitin and Mohan, Jayashree and Panwar, Ashish  and Kwatra, Nipun and Gulavani, Bhargav S and Ramjee, Ramachandran and Tumanov, Alexey},
          journal={Proceedings of The Seventh Annual Conference on Machine Learning and Systems, 2024, Santa Clara},
          year={2024}
        }
        ```
        """
    )



import logging

import pandas as pd

from vidur.config import TraceRequestIntervalGeneratorConfig
from vidur.request_generator.base_request_interval_generator import (
    BaseRequestIntervalGenerator,
)

logger = logging.getLogger(__name__)


class TraceRequestIntervalGenerator(BaseRequestIntervalGenerator):
    """
    Reads a trace csv file containing request arrival time, its prompt and completion token values to generate
    inter-request times, number of tokens.
    """

    def __init__(self, config: TraceRequestIntervalGeneratorConfig):
        super().__init__(config)

        # load into a pd dataframe
        self.trace_df = pd.read_csv(config.trace_file)

        self.trace_df["arrival_time"] = pd.to_datetime(self.trace_df["arrival_time"])
        # restrict trace_df to be a subset of rows that have the same date
        self.trace_df = self.trace_df[
            (self.trace_df["arrival_time"] > config.start_time)
            & (self.trace_df["arrival_time"] < config.end_time)
        ]

        # change back to seconds
        self.trace_df["arrival_time"] = (
            self.trace_df["arrival_time"] - self.trace_df["arrival_time"].min()
        ) // pd.Timedelta("1s")

        # rescale the time to change QPS
        self.trace_df["arrival_time"] = (
            self.trace_df["arrival_time"] * config.time_scale_factor
        )

        # compute the inter-request time
        self.trace_df["inter_request_time"] = self.trace_df["arrival_time"].diff()

        self.next_request_idx = 1

        logger.info(
            f"Loaded interval trace file {config.trace_file} with {len(self.trace_df)} requests"
        )

    def get_next_inter_request_time(self) -> float:
        if self.next_request_idx >= len(self.trace_df):
            return None

        inter_request_time = self.trace_df.iloc[self.next_request_idx][
            "inter_request_time"
        ]
        self.next_request_idx += 1

        return inter_request_time



import plotly.express as px
import randomname
import streamlit as st

from vidur.config_optimizer.analyzer.constants import AXIS_COLS, PRETTY_NAMES


def hide_anchor_link():
    st.markdown(
        """
        <style>
        .st-emotion-cache-15zrgzn.e1nzilvr3 {display: none}
        </style>
        """,
        unsafe_allow_html=True,
    )


def add_small_divider():
    st.markdown(
        """<hr style="height:1px;border:none;color:#333;background-color:#333;margin-top: 3px;margin-bottom: 20px;" /> """,
        unsafe_allow_html=True,
    )


def get_model_trace_names(subset_dfs):
    model_trace_names = list(subset_dfs.keys())
    model_names = set([model for model, _ in model_trace_names])
    trace_names = set([trace for _, trace in model_trace_names])
    return model_names, trace_names


def add_model_trace_selector(subset_dfs):
    model_names, trace_names = get_model_trace_names(subset_dfs)

    col1, col2 = st.columns(2)

    with col1:
        model_select_box = st.selectbox("Model:", model_names)
    with col2:
        trace_select_box = st.selectbox("Trace:", trace_names)

    return model_select_box, trace_select_box


def get_config_name(config_row):
    axis_cols = ["Model", "Trace"] + list(AXIS_COLS.values())
    axis_dict = {col: config_row[col] for col in axis_cols}
    randomname.util.rng.seed(hash(str(axis_dict)))
    return randomname.core.generate(
        randomname.util.prefix("a", ("music_theory", "sound", "size")),
        randomname.util.prefix("n", ("cats", "food", "apex_predators")),
        sep="-",
    )


def add_advanced_filters(df):
    with st.expander("Advanced Filters"):
        col1, col2 = st.columns(2)

        config_filters = {}
        for i, axis_col in enumerate(AXIS_COLS.values()):
            col = col1 if i % 2 == 0 else col2
            with col:
                config_filters[axis_col] = st.multiselect(
                    axis_col, df[axis_col].unique(), key=f"{axis_col}_selector"
                )

    # apply filters
    filtered_df = df
    for axis_col, values in config_filters.items():
        if len(values) == 0:
            continue
        filtered_df = filtered_df[filtered_df[axis_col].isin(values)]

    return filtered_df


def plot_cdf(config_df, y, y_name, color):
    chart_col, _ = st.columns([2.5, 1])
    # split the cdf_series from str to float list
    config_df["cdf_series"] = config_df[y].apply(
        lambda x: list(map(float, x[1:-1].split(",")))
    )
    # create a separate row for each cdf value and add a new column cdf_x
    config_df = config_df.explode("cdf_series").reset_index(drop=True)
    config_df["cdf_x"] = config_df.groupby(
        ["Name", "poisson_request_interval_generator_qps"]
    ).cumcount()

    with chart_col:
        fig = px.line(
            config_df,
            x="cdf_x",
            y="cdf_series",
            color=color,
            labels={"cdf_series": y_name, "cdf_x": "CDF"},
            title=f"{y_name}",
        )

        fig.update_layout(
            font=dict(size=14),
            legend=dict(
                title_font=dict(size=16),
                font=dict(size=14),
            ),
            xaxis_title=y_name,
            yaxis_title="CDF",
        )

        st.plotly_chart(fig, use_container_width=True)


def write_best_config(
    bottleneck_analyzer,
    df,
    metric_1,
    percentile_1,
    slo_1,
    metric_2,
    percentile_2,
    slo_2,
    use_qps=False,
):
    metric_1_col = f"{metric_1}_{percentile_1}%"
    metric_2_col = f"{metric_2}_{percentile_2}%"

    target_col = "QPS" if use_qps else "QPS per Dollar"

    slo_compliant_df = df[(df[metric_1_col] <= slo_1) & (df[metric_2_col] <= slo_2)]
    best_config = slo_compliant_df.sort_values(target_col, ascending=False)

    if len(best_config) == 0 or best_config[target_col].iloc[0] == 0:
        st.markdown("No config satisfies the SLOs")
        return

    best_config = best_config.iloc[0]

    best_point_str = ""
    for _, axis_name in AXIS_COLS.items():
        if (
            axis_name == "Sarathi Chunk Size"
            and best_config["Scheduler"] != "Sarathi-Serve"
        ):
            continue
        value = best_config[axis_name]
        value = str(value) if not value in PRETTY_NAMES else PRETTY_NAMES[value]
        best_point_str += f"**{axis_name}**: {value}, "

    best_point_str = best_point_str[:-2]
    best_point_str += f"\n\n"

    if use_qps:
        best_point_str += f"**QPS**: {best_config['QPS']:.2f}, "
    best_point_str += f"**QPS per Dollar**: {best_config['QPS per Dollar']:.2f}, "
    best_point_str += f"**{metric_1_col}**: {best_config[metric_1_col]:.2f}, "
    best_point_str += f"**{metric_2_col}**: {best_config[metric_2_col]:.2f}"

    best_point_str += f"\n\n**MFU**: {best_config['mfu_mean']:.2f}, "
    best_point_str += f"**Memory Usage**: {best_config['memory_usage_mean']:.2f}%, "
    best_point_str += f"**Busy Time**: {best_config['busy_time_percent_mean']:.2f}%"

    bottleneck_message = bottleneck_analyzer.analyze(best_config)
    best_point_str += f"\n\n**Bottleneck**: {bottleneck_message}"

    st.markdown(f"#### Best Config \n\n{best_point_str}")


def convert_config_row_to_comparison_point(config_row):
    COMPARE_COLS = ["Model", "Trace"] + list(AXIS_COLS.values())
    return {col: config_row[col] for col in COMPARE_COLS}



# Test the creation of cpp entities, configs and execution time predictor

import pytest
from vidur._native.entities import Batch as BatchC, KVParallelBatch as KVParallelBatchC, ExecutionTime as ExecutionTimeC
from vidur._native.execution_time_predictor import ExecutionTimePredictor as ExecutionTimePredictorC
from vidur._native.config import ExecutionTimePredictorConfig as ExecutionTimePredictorConfigC, ReplicaConfig as ReplicaConfigC, ModelConfig as ModelConfigC

def test_batch_creation():
    # Test basic batch creation
    batch = BatchC(
        replica_id=0,
        num_requests=2,
        num_q_tokens=[10, 15],
        num_kv_tokens=[20, 25],
        num_active_kvp_groups=[1, 2],
        kvp_group_id=0
    )
    
    # Test properties
    assert batch.replica_id == 0
    assert batch.num_requests == 2
    assert list(batch.num_q_tokens) == [10, 15]
    assert list(batch.num_kv_tokens) == [20, 25]
    assert list(batch.num_active_kvp_groups) == [1, 2]
    assert batch.kvp_group_id == 0
    assert batch.total_num_q_tokens == 25  # 10 + 15
    assert batch.total_num_kv_tokens == 45  # 20 + 25

def test_kv_parallel_batch():
    # Create some batches
    batch1 = BatchC(0, 2, [10, 15], [20, 25], [1, 2], 0)
    batch2 = BatchC(0, 2, [12, 18], [22, 28], [1, 2], 1)
    
    # Create KVParallelBatch
    kvp_batch = KVParallelBatchC(
        replica_id=0,
        kvp_group_ids=[0, 1],
        batches=[batch1, batch2]
    )
    
    # Test properties
    assert kvp_batch.replica_id == 0
    batch_map = kvp_batch.batch_mapping
    assert len(batch_map) == 2
    assert 0 in batch_map
    assert 1 in batch_map

def test_execution_time():
    # Test execution time creation
    exec_time = ExecutionTimeC(
        num_layers_per_pipeline_stage=4,
        attention_rope_execution_time=1.0,
        attention_kv_cache_save_execution_time=1.0,
        attention_decode_execution_time=2.0,
        attention_prefill_execution_time=2.0,
        attention_layer_pre_proj_execution_time=1.0,
        attention_layer_post_proj_execution_time=1.0,
        mlp_layer_up_proj_execution_time=1.0,
        mlp_layer_down_proj_execution_time=1.0,
        mlp_layer_act_execution_time=1.0,
        attn_norm_time=1.0,
        mlp_norm_time=1.0,
        add_time=1.0,
        tensor_parallel_communication_time=0.5,
        pipeline_parallel_communication_time=0.5,
    )
    
    # Test methods
    assert exec_time.model_time > 0
    assert exec_time.total_time > 0

def test_replica_config():
    # Test config creation
    config = ReplicaConfigC(
        num_pipeline_stages=2,
        tensor_parallel_size=1,
        kv_parallel_size=1
    )
    
    # Test properties
    assert config.num_pipeline_stages == 2
    assert config.tensor_parallel_size == 1
    assert config.kv_parallel_size == 1

def test_model_config():
    config = ModelConfigC(
        num_layers=8,
        num_q_heads=16,
        num_kv_heads=16,
        embedding_dim=1024,
        mlp_hidden_dim=4096,
        max_model_len=4096,
        use_gated_mlp=True,
        use_bias=True,
        use_qkv_bias=True,
        post_attn_norm=True,
        vocab_size=10000
    )
    
    # Test properties
    assert config.num_layers == 8
    assert config.num_q_heads == 16
    assert config.num_kv_heads == 16
    assert config.embedding_dim == 1024
    assert config.mlp_hidden_dim == 4096
    assert config.max_model_len == 4096
    assert config.use_gated_mlp == True
    assert config.use_bias == True
    assert config.use_qkv_bias == True
    assert config.post_attn_norm == True
    assert config.vocab_size == 10000

def test_execution_time_predictor_config():
    # Test config creation
    config = ExecutionTimePredictorConfigC(
        kv_cache_prediction_granularity=256,
        prediction_max_prefill_chunk_size=4096,
        prediction_max_batch_size=128,
        prediction_max_tokens_per_request=2 * 1024 * 1024,
        attention_decode_batching_overhead_fraction=0.1,
        nccl_cpu_launch_overhead_ms=0.02,
        nccl_cpu_skew_overhead_per_device_ms=0.0,
        use_native_execution_time_predictor=True,
        disable_kvp_communication=True
    )
    
    # Test properties
    assert config.kv_cache_prediction_granularity == 256
    assert config.prediction_max_prefill_chunk_size == 4096
    assert config.prediction_max_batch_size == 128
    assert config.prediction_max_tokens_per_request == 2 * 1024 * 1024
    assert config.attention_decode_batching_overhead_fraction == 0.1
    assert config.nccl_cpu_launch_overhead_ms == 0.02
    assert config.nccl_cpu_skew_overhead_per_device_ms == 0.0
    assert config.use_native_execution_time_predictor
    assert config.disable_kvp_communication

def test_execution_time_predictor():
    # Create configs using native implementations
    model_config = ModelConfigC(
        num_layers=8,
        num_q_heads=16,
        num_kv_heads=16,
        embedding_dim=1024,
        mlp_hidden_dim=4096,
        max_model_len=4096,
        use_gated_mlp=True,
        use_bias=True,
        use_qkv_bias=True,
        post_attn_norm=True,
        vocab_size=10000
    )
    
    replica_config = ReplicaConfigC(
        num_pipeline_stages=2,
        tensor_parallel_size=1,
        kv_parallel_size=1
    )
    
    predictor_config = ExecutionTimePredictorConfigC(
        kv_cache_prediction_granularity=256,
        prediction_max_prefill_chunk_size=4096,
        prediction_max_batch_size=128,
        prediction_max_tokens_per_request=2 * 1024 * 1024,
        attention_decode_batching_overhead_fraction=0.1,
        nccl_cpu_launch_overhead_ms=0.02,
        nccl_cpu_skew_overhead_per_device_ms=0.0,
        use_native_execution_time_predictor=True,
        disable_kvp_communication=True
    )
    
    # Create some sample prediction data
    prediction_ops = ["attention", "mlp"]
    prediction_keys = [[(1, 1), (2, 2)], [(3, 3), (4, 4)]]
    prediction_values = [[0.1, 0.2], [0.3, 0.4]]
    
    # Create predictor
    predictor = ExecutionTimePredictorC(
        predictor_config,
        replica_config,
        model_config,
        prediction_ops,
        prediction_keys,
        prediction_values
    )
    
    # Test prediction with a batch
    batch = BatchC(0, 2, [10, 15], [20, 25], [1, 2], 0)
    exec_time = predictor.get_execution_time_batch(batch, pipeline_stage=0)
    assert isinstance(exec_time, ExecutionTimeC)
    
    # Test prediction with KVParallelBatch
    batch1 = BatchC(0, 2, [10, 15], [20, 25], [1, 2], 0)
    batch2 = BatchC(0, 2, [12, 18], [22, 28], [1, 2], 1)
    kvp_batch = KVParallelBatchC(0, [0, 1], [batch1, batch2])
    exec_time = predictor.get_execution_time_kv_parallel_batch(kvp_batch, pipeline_stage=0)
    assert isinstance(exec_time, ExecutionTimeC)



from vidur.request_generator.synthetic_request_generator import (
    SyntheticRequestGenerator,
)
from vidur.request_generator.trace_replay_request_generator import (
    TraceReplayRequestGenerator,
)
from vidur.types import RequestGeneratorType
from vidur.utils.base_registry import BaseRegistry


class RequestGeneratorRegistry(BaseRegistry):
    pass


RequestGeneratorRegistry.register(
    RequestGeneratorType.SYNTHETIC, SyntheticRequestGenerator
)
RequestGeneratorRegistry.register(
    RequestGeneratorType.TRACE_REPLAY, TraceReplayRequestGenerator
)



from abc import ABC
from dataclasses import dataclass
from typing import Any

from vidur.config.utils import get_all_subclasses


@dataclass
class BaseFixedConfig(ABC):

    @classmethod
    def create_from_type(cls, type_: Any) -> Any:
        for subclass in get_all_subclasses(cls):
            if subclass.get_type() == type_:
                return subclass()
        raise ValueError(f"[{cls.__name__}] Invalid type: {type_}")

    @classmethod
    def create_from_name(cls, name: str) -> Any:
        for subclass in get_all_subclasses(cls):
            if subclass.get_name() == name:
                return subclass()
        raise ValueError(f"[{cls.__name__}] Invalid name: {name}")

    @classmethod
    def create_from_type_string(cls, type_str: str) -> Any:
        for subclass in get_all_subclasses(cls):
            if str(subclass.get_type()) == type_str:
                return subclass()
        raise ValueError(f"[{cls.__name__}] Invalid type string: {type_str}")



import json

from vidur.config import BaseRequestGeneratorConfig, ClusterConfig, MetricsConfig
from vidur.entities.base_entity import BaseEntity
from vidur.entities.replica import Replica
from vidur.logger import init_logger

logger = init_logger(__name__)


class Cluster(BaseEntity):
    def __init__(
        self,
        cluster_config: ClusterConfig,
        metrics_config: MetricsConfig,
        generator_config: BaseRequestGeneratorConfig,
    ) -> None:
        self._id = Cluster.generate_id()
        self._config = cluster_config

        # get metrics config
        self._output_dir = metrics_config.output_dir

        # Init replica object handles
        self._replicas = {}

        for _ in range(self._config.num_replicas):
            replica = Replica(self._config.replica_config, generator_config)
            self._replicas[replica.id] = replica

        if metrics_config.write_json_trace:
            self._write_cluster_info_to_file()

    @property
    def replicas(self):
        return self._replicas

    def to_dict(self) -> dict:
        return {
            "id": self._id,
            "num_replicas": len(self._replicas),
        }

    def _write_cluster_info_to_file(self) -> None:
        replica_dicts = [replica.to_dict() for replica in self._replicas.values()]
        cluster_info = {"replicas": replica_dicts}

        cluster_file = f"{self._output_dir}/cluster.json"
        with open(cluster_file, "w") as f:
            json.dump(cluster_info, f)



from typing import Tuple

from vidur.config import ZipfRequestLengthGeneratorConfig
from vidur.request_generator.base_request_length_generator import (
    BaseRequestLengthGenerator,
)
from vidur.utils.zipf_generator import ZipfGenerator


class ZipfRequestLengthGenerator(BaseRequestLengthGenerator):

    def __init__(self, config: ZipfRequestLengthGeneratorConfig):
        super().__init__(config)

        self.zipf_generator = ZipfGenerator(
            config.min_tokens,
            config.max_tokens,
            config.theta,
            config.scramble,
            config.seed,
        )

    def get_next_num_tokens(self) -> Tuple[float, float]:
        total_tokens = self.zipf_generator.next()

        decode_tokens = total_tokens / (1 + self.config.prefill_to_decode_ratio)
        prefill_tokens = total_tokens - decode_tokens

        return prefill_tokens, decode_tokens



from vidur.types.base_int_enum import BaseIntEnum


class ReplicaSchedulerType(BaseIntEnum):
    FASTER_TRANSFORMER = 1
    ORCA = 2
    SARATHI = 3
    VLLM = 4
    LIGHTLLM = 5
    MNEMOSYNE_FCFS_FIXED_CHUNK = 6
    MNEMOSYNE_FCFS = 7
    MNEMOSYNE_EDF = 8
    MNEMOSYNE_LRS = 9
    MNEMOSYNE_ST = 10



from typing import List

from vidur.entities.base_entity import BaseEntity
from vidur.entities.batch_stage import BatchStage
from vidur.entities.execution_time import ExecutionTime
from vidur.entities.request import Request
from vidur.logger import init_logger

logger = init_logger(__name__)


# a decorator which checks if the request has been scheduled
def check_scheduled(func):
    def wrapper(self, *args, **kwargs):
        if not self._scheduled:
            raise ValueError("Batch has not been scheduled yet")
        return func(self, *args, **kwargs)

    return wrapper


def check_completed(func):
    def wrapper(self, *args, **kwargs):
        if not self._completed:
            raise ValueError("Batch has not been scheduled yet")
        return func(self, *args, **kwargs)

    return wrapper


class Batch(BaseEntity):
    def __init__(
        self,
        replica_id: int,
        requests: List[Request],
        num_q_tokens: List[int],
        num_kv_tokens: List[int],
        num_active_kvp_groups: List[int] = None,
        last_kvp_group_ids: List[int] = None,
        kvp_group_id: int = 0,
    ) -> None:
        self._id = Batch.generate_id()
        self._replica_id = replica_id

        self._requests = requests
        self._num_q_tokens = num_q_tokens
        self._num_kv_tokens = num_kv_tokens
        self._num_active_kvp_groups = num_active_kvp_groups
        self._last_kvp_group_ids = last_kvp_group_ids
        self._kvp_group_id = kvp_group_id
        self._total_num_q_tokens = sum(num_q_tokens)
        self._total_num_kv_tokens = sum(num_kv_tokens)
        self._total_num_q_tokens_rounded = (self._total_num_q_tokens + 7) // 8 * 8
        self._num_prefill_tokens = sum(
            [
                (t if not r.is_prefill_complete else 0)
                for r, t in zip(self.requests, self._num_q_tokens)
            ]
        )

        self._scheduled_at = None
        self._completed_at = None
        self._scheduled = False
        self._completed = False

    @property
    def replica_id(self) -> int:
        return self._replica_id

    @property
    def creation_time(self) -> float:
        return self._creation_time

    @property
    def num_q_tokens(self) -> List[int]:
        return self._num_q_tokens

    @property
    def num_kv_tokens(self) -> List[int]:
        return self._num_kv_tokens

    @property
    def num_active_kvp_groups(self) -> List[int]:
        return self._num_active_kvp_groups

    @property
    def kvp_group_id(self) -> int:
        return self._kvp_group_id

    @property
    def last_kvp_group_ids(self) -> List[int]:
        return self._last_kvp_group_ids

    @property
    def total_num_q_tokens(self) -> int:
        return self._total_num_q_tokens

    @property
    def total_num_kv_tokens(self) -> int:
        return self._total_num_kv_tokens

    @property
    def total_num_q_tokens_rounded(self) -> int:
        return self._total_num_q_tokens_rounded

    @property
    def num_prefill_tokens(self) -> int:
        return self._num_prefill_tokens

    @property
    def num_decode_tokens(self) -> int:
        return self.total_num_q_tokens - self.num_prefill_tokens

    @property
    @check_scheduled
    def scheduled_at(self) -> float:
        return self._scheduled_at

    @property
    @check_completed
    def completed_at(self) -> float:
        return self._completed_at

    @property
    def completed(self) -> bool:
        return self._completed

    @property
    def scheduled(self) -> bool:
        return self._scheduled

    @property
    def size(self) -> int:
        return len(self._requests)

    @property
    def requests(self) -> List[Request]:
        return self._requests

    @property
    def request_ids(self) -> List[int]:
        return [request.id for request in self._requests]

    @property
    def have_all_requests_completed(self) -> bool:
        return all([request.completed for request in self._requests])

    def on_schedule(
        self,
        time: float,
    ) -> None:
        self._scheduled_at = time
        self._scheduled = True

        if not self._last_kvp_group_ids:
            for request in self._requests:
                request.on_batch_schedule(time)
            return

        for request, last_kvp_group_id in zip(self._requests, self._last_kvp_group_ids):
            if self._kvp_group_id != last_kvp_group_id:
                continue

            request.on_batch_schedule(time)

    def on_batch_end(self, time: float):
        self._completed = True
        self._completed_at = time

        if not self._last_kvp_group_ids:
            assert self._kvp_group_id == 0
            for request, num_q_tokens in zip(self._requests, self._num_q_tokens):
                request.on_batch_end(time, num_q_tokens)
            return

        for request, num_q_tokens, last_kvp_group_id in zip(
            self._requests, self._num_q_tokens, self._last_kvp_group_ids
        ):
            if self._kvp_group_id != last_kvp_group_id:
                continue

            request.on_batch_end(time, num_q_tokens)

    @property
    def preempted_requests(self) -> List[Request]:
        return [request for request in self._requests if request.preempted]

    @property
    def completed_requests(self) -> List[Request]:
        return [request for request in self._requests if request.completed]

    def to_batch_stage(
        self, stage_id: int, execution_time: ExecutionTime
    ) -> BatchStage:
        return BatchStage(
            self._id,
            self._replica_id,
            stage_id,
            execution_time,
            self._requests,
            self._num_q_tokens,
            self._num_kv_tokens,
            self._num_active_kvp_groups,
            self._last_kvp_group_ids,
            self._kvp_group_id,
        )

    def to_dict(self) -> dict:
        return {
            "id": self._id,
            "size": self.size,
            "replica_id": self._replica_id,
            "scheduled_at": self._scheduled_at,
            "completed_at": self._completed_at,
            "scheduled": self._scheduled,
            "request_ids": self.request_ids,
            "num_q_tokens": self.num_q_tokens,
            "num_kv_tokens": self.num_kv_tokens,
            "num_prefill_tokens": self.num_prefill_tokens,
            "num_decode_tokens": self.num_decode_tokens,
        }



import os
from functools import reduce
from typing import Dict, List, Union

import pandas as pd
import plotly_express as px
import wandb

from vidur.config import ClusterConfig, MetricsConfig, SimulationConfig
from vidur.entities import (
    Batch,
    BatchStage,
    ExecutionTime,
    KVParallelBatch,
    KVParallelBatchStage,
    Request,
)
from vidur.logger import init_logger
from vidur.metrics.cdf_sketch import CDFSketch
from vidur.metrics.constants import (
    BatchMetricsCountDistribution,
    BatchMetricsTimeDistribution,
    OperationMetrics,
    RequestCompletionMetricsTimeSeries,
    RequestMetricsHistogram,
    RequestMetricsTimeDistributions,
    TokenCompletionMetricsTimeSeries,
    TokenMetricsTimeDistribution,
)
from vidur.metrics.data_series import DataSeries
from vidur.metrics.series_average_meter import SeriesAverageMeter
from vidur.utils.mfu_calculator import MFUCalculator

logger = init_logger(__name__)


def if_write_metrics(func):
    def wrapper(self, *args, **kwargs):
        if self._config.write_metrics:
            return func(self, *args, **kwargs)

    return wrapper


REQUEST_ID_STR = "Request Id"
COUNT_STR = "Count"
TIME_STR = "Time (sec)"
BATCH_ID_STR = "Batch Id"
MEMORY_USAGE_STR = "Memory Usage (%)"
BUSY_TIME_PERCENT = "Busy Time (%)"
UTILIZATION_STR = "Utilization (%)"
OPERATION_STR = "Operation"
TIME_STR_MS = "Time (ms)"


class MetricsStore:

    def __init__(self, simulation_config: SimulationConfig) -> None:
        self._simulation_config = simulation_config
        self._config = simulation_config.metrics_config

        self._last_request_arrived_at = None

        # copy config
        self._num_replicas = simulation_config.cluster_config.num_replicas
        self._num_pipeline_stages = (
            simulation_config.cluster_config.replica_config.num_pipeline_stages
        )

        # Initialise request metrics
        self._request_metrics_time_distributions: Dict[
            RequestMetricsTimeDistributions, DataSeries
        ] = {}
        for metric_name in RequestMetricsTimeDistributions:
            self._request_metrics_time_distributions[metric_name] = DataSeries(
                REQUEST_ID_STR,
                metric_name.value,
                self._config.subsamples,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )

        self._token_metrics_time_distribution: Dict[
            TokenMetricsTimeDistribution, DataSeries
        ] = {}
        for metric_name in TokenMetricsTimeDistribution:
            self._token_metrics_time_distribution[metric_name] = CDFSketch(
                metric_name.value,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )

        self._request_metrics_histogram: Dict[RequestMetricsHistogram, DataSeries] = {}
        for metric_name in RequestMetricsHistogram:
            self._request_metrics_histogram[metric_name] = DataSeries(
                REQUEST_ID_STR,
                metric_name.value,
                self._config.subsamples,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )

        # Initialise batch metrics
        self._batch_metrics_count_distribution: Dict[
            BatchMetricsCountDistribution, CDFSketch
        ] = {}
        self._batch_metrics_count_distribution_per_batch: Dict[
            BatchMetricsCountDistribution, DataSeries
        ] = {}
        for metric_name in BatchMetricsCountDistribution:
            self._batch_metrics_count_distribution[metric_name] = CDFSketch(
                metric_name.value,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )
            self._batch_metrics_count_distribution_per_batch[metric_name] = DataSeries(
                BATCH_ID_STR,
                metric_name.value,
                self._config.subsamples,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )

        self._batch_metrics_time_distribution: Dict[
            BatchMetricsTimeDistribution, CDFSketch
        ] = {}
        self._batch_metrics_time_distribution_per_batch: Dict[
            BatchMetricsTimeDistribution, DataSeries
        ] = {}
        for metric_name in BatchMetricsTimeDistribution:
            self._batch_metrics_time_distribution[metric_name] = CDFSketch(
                metric_name.value,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )
            self._batch_metrics_time_distribution_per_batch[metric_name] = DataSeries(
                BATCH_ID_STR,
                metric_name.value,
                self._config.subsamples,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )

        # Initialise completion metrics
        self._request_completion_metrics_time_series: Dict[
            RequestCompletionMetricsTimeSeries, DataSeries
        ] = {}
        for metric_name in RequestCompletionMetricsTimeSeries:
            self._request_completion_metrics_time_series[metric_name] = DataSeries(
                TIME_STR,
                metric_name.value,
                self._config.subsamples,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )
        self._token_completion_metrics_time_series: Dict[
            TokenCompletionMetricsTimeSeries, DataSeries
        ] = {}
        for metric_name in TokenCompletionMetricsTimeSeries:
            self._token_completion_metrics_time_series[metric_name] = DataSeries(
                TIME_STR,
                metric_name.value,
                self._config.subsamples,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )

        # Initialise operation metrics
        self._operation_metrics: Dict[OperationMetrics, CDFSketch] = {}
        self._operation_metrics_per_batch: Dict[OperationMetrics, DataSeries] = {}
        for metric_name in OperationMetrics:
            self._operation_metrics[metric_name] = CDFSketch(
                metric_name.value,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )
            self._operation_metrics_per_batch[metric_name] = DataSeries(
                BATCH_ID_STR,
                metric_name.value,
                self._config.subsamples,
                self._config.save_table_to_wandb,
                self._config.store_plots,
            )

        # per replica metrics
        self._replica_memory_usage = []
        # per replica stage metrics
        self._replica_busy_time = []
        self._replica_mfu = []
        self._mfu_calculator = MFUCalculator(
            simulation_config.cluster_config.replica_config
        )

        for replica_idx in range(self._num_replicas):
            self._replica_memory_usage.append(
                SeriesAverageMeter(
                    TIME_STR,
                    MEMORY_USAGE_STR,
                    self._config.save_table_to_wandb,
                )
            )
            self._replica_memory_usage[replica_idx].put(0, 0)

            self._replica_busy_time.append([])
            self._replica_mfu.append([])

            for stage_idx in range(self._num_pipeline_stages):
                self._replica_busy_time[replica_idx].append(
                    SeriesAverageMeter(
                        TIME_STR,
                        BUSY_TIME_PERCENT,
                        save_table_to_wandb=self._config.save_table_to_wandb,
                    )
                )
                self._replica_busy_time[replica_idx][stage_idx].put(0, 0)

                self._replica_mfu[replica_idx].append(
                    SeriesAverageMeter(
                        TIME_STR,
                        UTILIZATION_STR,
                        save_table_to_wandb=self._config.save_table_to_wandb,
                    )
                )
                self._replica_mfu[replica_idx][stage_idx].put(0, 0)

        self._init_wandb()

    def _init_wandb(self):
        if (
            not self._config.write_metrics
            or not self._config.wandb_project
            or not self._config.wandb_group
        ):
            return

        wandb.init(
            project=self._config.wandb_project,
            group=self._config.wandb_group,
            name=self._config.wandb_run_name,
            config=self._simulation_config.to_dict(),
        )

    def _save_as_csv(
        self,
        dataseries_list: List[DataSeries],
        key_to_join: str,
        base_path: str,
        file_name: str,
    ):
        os.makedirs(base_path, exist_ok=True)

        merged_df = reduce(
            lambda left, right: pd.merge(left, right, on=[key_to_join], how="outer"),
            [dataseries._to_df() for dataseries in dataseries_list],
        )
        merged_df.to_csv(f"{base_path}/{file_name}.csv", index=False)
        if wandb.run and self._config.save_table_to_wandb:
            wand_table = wandb.Table(dataframe=merged_df)
            wandb.log({f"{file_name}_table": wand_table}, step=0)

    def _store_bar_plot(
        self,
        base_path: str,
        plot_name: str,
        x_label: str,
        y_label: str,
        data: Dict[str, float],
    ):
        if wandb.run:
            wandb.log(
                {
                    plot_name: wandb.plot.bar(
                        wandb.Table(
                            dataframe=pd.DataFrame(
                                data=data.items(), columns=[x_label, y_label]
                            )
                        ),
                        x_label,
                        y_label,
                        title=plot_name,
                    )
                },
                step=0,
            )
        if self._config.store_plots:
            fig = px.bar(
                x=list(data.keys()),
                y=list(data.values()),
                labels={"x": x_label, "y": y_label},
            )
            fig.write_image(f"{base_path}/{plot_name}.png")

    def _store_operation_metrics(self, base_plot_path: str):
        if not self._config.store_operation_metrics:
            return

        total_operation_runtimes: Dict[str, float] = {}

        total_operation_runtimes["model_execution_e2e"] = 0
        for dataseries in self._operation_metrics.values():
            dataseries.plot_cdf(
                base_plot_path, f"{dataseries._metric_name}_execution_time", TIME_STR_MS
            )
            total_operation_runtimes[dataseries._metric_name] = dataseries.sum
            total_operation_runtimes["model_execution_e2e"] += dataseries.sum

        for dataseries in self._cpu_operation_metrics.values():
            dataseries.plot_cdf(
                base_plot_path, f"{dataseries._metric_name}_execution_time", TIME_STR_MS
            )
            total_operation_runtimes[dataseries._metric_name] = dataseries.sum

        self._store_bar_plot(
            base_plot_path,
            "total_operation_runtimes",
            OPERATION_STR,
            TIME_STR_MS,
            total_operation_runtimes,
        )

        if not self._config.keep_individual_batch_metrics:
            return

        for dataseries in self._operation_metrics_per_batch.values():
            dataseries.consolidate()
            dataseries.plot_step(
                base_plot_path,
                f"{dataseries._metric_name}_per_batch",
                y_axis_label=TIME_STR_MS,
                y_cumsum=False,
            )
        operations_dataseries_list = list(self._operation_metrics_per_batch.values())
        self._save_as_csv(
            dataseries_list=operations_dataseries_list,
            key_to_join=BATCH_ID_STR,
            base_path=self._config.output_dir,
            file_name="operation_metrics",
        )

    def _store_request_metrics(self, base_plot_path: str):
        if not self._config.store_request_metrics:
            return

        all_request_metrics = list(
            self._request_metrics_time_distributions.values()
        ) + list(self._request_metrics_histogram.values())

        self._save_as_csv(
            dataseries_list=all_request_metrics,
            key_to_join=REQUEST_ID_STR,
            base_path=self._config.output_dir,
            file_name="request_metrics",
        )

        for dataseries in self._request_metrics_histogram.values():
            dataseries.plot_histogram(base_plot_path, dataseries._y_name)

        for dataseries in self._request_metrics_time_distributions.values():
            dataseries.plot_cdf(base_plot_path, dataseries._y_name, TIME_STR)

    def _store_batch_metrics(self, base_plot_path: str):
        if not self._config.store_batch_metrics:
            return

        for dataseries in self._batch_metrics_time_distribution.values():
            y_axis_label = (
                TIME_STR_MS
                if "model_execution" in dataseries._metric_name
                else TIME_STR
            )
            dataseries.plot_cdf(base_plot_path, dataseries._metric_name, y_axis_label)

        for dataseries in self._batch_metrics_count_distribution.values():
            dataseries.plot_cdf(base_plot_path, dataseries._metric_name, COUNT_STR)

        if not self._config.keep_individual_batch_metrics:
            return

        for dataseries in self._batch_metrics_time_distribution_per_batch.values():
            y_axis_label = (
                TIME_STR_MS
                if "model_execution" in dataseries._metric_name
                else TIME_STR
            )
            dataseries.plot_step(
                base_plot_path,
                f"{dataseries._metric_name}_per_batch",
                y_axis_label=y_axis_label,
                y_cumsum=False,
            ),

        for dataseries in self._batch_metrics_count_distribution_per_batch.values():
            dataseries.plot_step(
                base_plot_path,
                f"{dataseries._metric_name}_per_batch",
                y_axis_label=COUNT_STR,
                y_cumsum=False,
            ),

        all_batch_metrics = list(
            self._batch_metrics_count_distribution_per_batch.values()
        ) + list(self._batch_metrics_time_distribution_per_batch.values())

        self._save_as_csv(
            dataseries_list=all_batch_metrics,
            key_to_join=BATCH_ID_STR,
            base_path=self._config.output_dir,
            file_name="batch_metrics",
        )

    def _store_completion_metrics(self, base_plot_path: str):
        if self._config.store_request_metrics:
            for dataseries in self._request_completion_metrics_time_series.values():
                dataseries.plot_step(
                    base_plot_path, f"{dataseries._y_name}_time_series", COUNT_STR
                )

        if not self._config.store_token_completion_metrics:
            return

        for dataseries in self._token_metrics_time_distribution.values():
            dataseries.plot_cdf(base_plot_path, dataseries._metric_name, TIME_STR)

        for dataseries in self._token_completion_metrics_time_series.values():
            dataseries.plot_step(
                base_plot_path, f"{dataseries._y_name}_time_series", COUNT_STR
            )

    def _store_utilization_metrics(self, base_plot_path: str):
        if not self._config.store_utilization_metrics:
            return

        for replica_idx in range(self._num_replicas):
            self._replica_memory_usage[replica_idx].print_stats(
                f"replica_{replica_idx + 1}_memory_usage", base_plot_path
            )
            for stage_idx in range(self._num_pipeline_stages):
                self._replica_busy_time[replica_idx][stage_idx].print_stats(
                    f"replica_{replica_idx + 1}_stage_{stage_idx + 1}_busy_time_percent",
                    base_plot_path,
                )
                self._replica_mfu[replica_idx][stage_idx].print_stats(
                    f"replica_{replica_idx + 1}_stage_{stage_idx + 1}_mfu",
                    base_plot_path,
                )

    @if_write_metrics
    def plot(self) -> None:
        dir_plot_path = f"{self._config.output_dir}/plots"
        os.makedirs(dir_plot_path, exist_ok=True)

        self._store_request_metrics(dir_plot_path)
        self._store_batch_metrics(dir_plot_path)
        self._store_completion_metrics(dir_plot_path)
        self._store_operation_metrics(dir_plot_path)
        self._store_utilization_metrics(dir_plot_path)

    @if_write_metrics
    def on_request_arrival(self, time: float, request: Request) -> None:
        if not self._config.store_request_metrics:
            return

        self._request_completion_metrics_time_series[
            RequestCompletionMetricsTimeSeries.REQUEST_ARRIVAL
        ].put(time, 1)

        self._request_metrics_histogram[RequestMetricsHistogram.REQUEST_NUM_TOKENS].put(
            request.id, request.total_tokens
        )
        self._request_metrics_histogram[
            RequestMetricsHistogram.REQUEST_PREFILL_TOKENS
        ].put(request.id, request.num_prefill_tokens)
        self._request_metrics_histogram[
            RequestMetricsHistogram.REQUEST_DECODE_TOKENS
        ].put(request.id, request.num_decode_tokens)
        self._request_metrics_histogram[RequestMetricsHistogram.REQUEST_PD_RATIO].put(
            request.id, request.pd_ratio
        )
        if self._last_request_arrived_at is not None:
            self._request_metrics_histogram[
                RequestMetricsHistogram.REQUEST_INTER_ARRIVAL_DELAY
            ].put(request.id, request.arrived_at - self._last_request_arrived_at)
        self._last_request_arrived_at = request.arrived_at

    @if_write_metrics
    def _on_request_end(self, time: float, request: Request) -> None:
        if not self._config.store_request_metrics:
            return

        self._request_completion_metrics_time_series[
            RequestCompletionMetricsTimeSeries.REQUEST_COMPLETION
        ].put(request.completed_at, 1)

        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_E2E_TIME
        ].put(request.id, request.e2e_time)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_E2E_TIME_NORMALIZED
        ].put(request.id, request.e2e_time_normalized)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_EXECUTION_TIME
        ].put(request.id, request.execution_time)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_EXECUTION_TIME_NORMALIZED
        ].put(request.id, request.execution_time_normalized)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_MODEL_EXECUTION_TIME
        ].put(request.id, request.model_execution_time)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_MODEL_EXECUTION_TIME_NORMALIZED
        ].put(request.id, request.model_execution_time_normalized)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_PREEMPTION_TIME
        ].put(request.id, request.preempted_time)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_SCHEDULING_DELAY
        ].put(request.id, request.scheduling_delay)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_EXECUTION_PLUS_PREEMPTION_TIME
        ].put(request.id, request.execution_time + request.preempted_time)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.REQUEST_EXECUTION_PLUS_PREEMPTION_TIME_NORMALIZED
        ].put(
            request.id,
            (request.execution_time + request.preempted_time)
            / request.num_decode_tokens,
        )
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.PREFILL_TIME_E2E
        ].put(request.id, request.prefill_completed_at - request.arrived_at)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.PREFILL_TIME_EXECUTION_PLUS_PREEMPTION
        ].put(request.id, request.prefill_completed_at - request.scheduled_at)
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.PREFILL_TIME_EXECUTION_PLUS_PREEMPTION_NORMALIZED
        ].put(
            request.id,
            (request.prefill_completed_at - request.scheduled_at)
            / request.num_prefill_tokens,
        )
        self._request_metrics_time_distributions[
            RequestMetricsTimeDistributions.DECODE_TIME_EXECUTION_PLUS_PREEMPTION_NORMALIZED
        ].put(
            request.id,
            (request.completed_at - request.prefill_completed_at)
            / request.num_decode_tokens,
        )

        self._request_metrics_histogram[
            RequestMetricsHistogram.REQUEST_NUM_RESTARTS
        ].put(request.id, request.num_restarts)

    def _update_per_token_execution_times(
        self, time: float, request: Request, batch: Union[Batch, KVParallelBatch]
    ) -> None:
        # if prefill has just finished in this iteration, update the prefill completion time series
        if (
            time == request.prefill_completed_at
            and self._config.store_token_completion_metrics
        ):
            self._token_completion_metrics_time_series[
                TokenCompletionMetricsTimeSeries.PREFILL_COMPLETIONS
            ].put(
                time,
                request.num_prefill_tokens,
            )

        # determine if this was prefill or decode token
        if not request.has_started_decode:
            return

        if not self._config.store_token_completion_metrics:
            return

        self._token_metrics_time_distribution[
            TokenMetricsTimeDistribution.DECODE_TOKEN_EXECUTION_PLUS_PREMPTION_TIME
        ].put(
            time - batch.scheduled_at + request.latest_iteration_scheduling_delay,
        )

        self._token_completion_metrics_time_series[
            TokenCompletionMetricsTimeSeries.DECODE_COMPLETIONS
        ].put(time, 1)

    def _push_metric(
        self, metric_name: OperationMetrics, batch_id: int, value: float
    ) -> None:
        if metric_name in OperationMetrics:
            self._operation_metrics[metric_name].put(value)
            self._operation_metrics_per_batch[metric_name].put(batch_id, value)
        elif metric_name in BatchMetricsTimeDistribution:
            self._batch_metrics_time_distribution[metric_name].put(value)
            self._batch_metrics_time_distribution_per_batch[metric_name].put(
                batch_id, value
            )
        elif metric_name in BatchMetricsCountDistribution:
            self._batch_metrics_count_distribution[metric_name].put(value)
            self._batch_metrics_count_distribution_per_batch[metric_name].put(
                batch_id, value
            )
        else:
            raise ValueError(f"Invalid metric name {metric_name}")

    @if_write_metrics
    def on_batch_end(
        self, time: float, batch: Batch, replica_id: int, memory_usage_percent: int
    ) -> None:
        if (
            self._config.min_batch_index and batch.id < self._config.min_batch_index
        ) or (self._config.max_batch_index and batch.id > self._config.max_batch_index):
            return

        if isinstance(batch, KVParallelBatch):
            for kvp_batch in batch.batch_mapping.values():
                self.on_batch_end(time, kvp_batch, replica_id, memory_usage_percent)
            return

        if batch.last_kvp_group_ids:
            for request, last_kvp_group_id in zip(
                batch.requests, batch.last_kvp_group_ids
            ):
                if batch._kvp_group_id != last_kvp_group_id:
                    continue

                self._update_per_token_execution_times(time, request, batch)
                if request.completed:
                    self._on_request_end(time, request)
        else:
            for request in batch.completed_requests:
                self._on_request_end(time, request)
            for request in batch.requests:
                self._update_per_token_execution_times(time, request, batch)

        if self._config.store_utilization_metrics:
            self._replica_memory_usage[replica_id - 1].put(time, memory_usage_percent)

        if not self._config.store_batch_metrics:
            return

        self._push_metric(
            BatchMetricsTimeDistribution.BATCH_EXECUTION_TIME,
            batch.id,
            time - batch.scheduled_at,
        )
        self._push_metric(
            BatchMetricsCountDistribution.BATCH_NUM_TOKENS,
            batch.id,
            batch.total_num_q_tokens,
        )
        self._push_metric(
            BatchMetricsCountDistribution.BATCH_NUM_PREFILL_TOKENS,
            batch.id,
            sum([(t if not t == 1 else 0) for t in batch.num_q_tokens]),
        )
        self._push_metric(
            BatchMetricsCountDistribution.BATCH_NUM_DECODE_TOKENS,
            batch.id,
            batch.num_decode_tokens,
        )
        self._push_metric(
            BatchMetricsCountDistribution.BATCH_SIZE, batch.id, batch.size
        )

    @if_write_metrics
    def on_replica_schedule(
        self, time: float, replica_id: int, memory_usage_percent: int
    ) -> None:
        if not self._config.store_utilization_metrics:
            return

        self._replica_memory_usage[replica_id - 1].put(time, memory_usage_percent)

    @if_write_metrics
    def on_replica_stage_schedule(
        self,
        time: float,
        replica_id: int,
        stage_id: int,
        batch_stage: Union[BatchStage, KVParallelBatchStage],
        execution_time: ExecutionTime,
    ) -> None:
        if not self._config.store_utilization_metrics:
            return

        if isinstance(batch_stage, KVParallelBatchStage):
            for kvp_stage in batch_stage.batch_stage_mapping.values():
                self.on_replica_stage_schedule(
                    time, replica_id, stage_id, kvp_stage, execution_time
                )
            return

        self._replica_busy_time[replica_id - 1][stage_id - 1].put(time, 100)
        mfu = self._mfu_calculator.get_mfu(batch_stage)
        self._replica_mfu[replica_id - 1][stage_id - 1].put(time, mfu)

        if not self._config.store_operation_metrics:
            return

        batch_id = batch_stage._batch_id
        for _ in range(execution_time.num_layers):
            self._push_metric(
                OperationMetrics.MLP_UP_PROJ,
                batch_id,
                execution_time.mlp_layer_up_proj_execution_time,
            )
            self._push_metric(
                OperationMetrics.MLP_ACTIVATION,
                batch_id,
                execution_time.mlp_layer_act_execution_time,
            )
            self._push_metric(
                OperationMetrics.MLP_DOWN_PROJ,
                batch_id,
                execution_time.mlp_layer_down_proj_execution_time,
            )
            self._push_metric(
                OperationMetrics.MLP_DOWN_PROJ_ALL_REDUCE,
                batch_id,
                execution_time.mlp_all_reduce_time,
            )
            self._push_metric(
                OperationMetrics.ATTN_PRE_PROJ,
                batch_id,
                execution_time.attention_pre_proj_time,
            )
            self._push_metric(
                OperationMetrics.ATTN_POST_PROJ,
                batch_id,
                execution_time.attention_post_proj_time,
            )
            self._push_metric(
                OperationMetrics.ATTN_POST_PROJ_ALL_REDUCE,
                batch_id,
                execution_time.attention_all_reduce_time,
            )

            if execution_time.attention_prefill_execution_time != 0:
                self._push_metric(
                    OperationMetrics.ATTN_PREFILL,
                    batch_id,
                    execution_time.attention_prefill_execution_time,
                )

            if execution_time.attention_decode_execution_time != 0:
                self._push_metric(
                    OperationMetrics.ATTN_DECODE,
                    batch_id,
                    execution_time.attention_decode_execution_time,
                )
            self._push_metric(
                OperationMetrics.ATTN_KV_CACHE_SAVE,
                batch_id,
                execution_time.attention_kv_cache_save_execution_time,
            )
            self._push_metric(
                OperationMetrics.ATTN_ROPE,
                batch_id,
                execution_time.attention_rope_execution_time,
            )
            self._push_metric(
                OperationMetrics.ADD, batch_id, execution_time.add_time * 2
            )
            self._push_metric(
                OperationMetrics.INPUT_LAYERNORM,
                batch_id,
                execution_time.attn_norm_time,
            )
            self._push_metric(
                OperationMetrics.POST_ATTENTION_LAYERNORM,
                batch_id,
                execution_time.mlp_norm_time,
            )

        self._push_metric(
            OperationMetrics.PIPELINE_SEND_RECV,
            batch_id,
            execution_time.pipeline_parallel_communication_time,
        )

    @if_write_metrics
    def on_batch_stage_end(
        self, time: float, batch_stage: BatchStage, replica_id: int, stage_id: int
    ) -> None:
        if not self._config.store_utilization_metrics:
            return
        self._replica_busy_time[replica_id - 1][stage_id - 1].put(time, 0)
        self._replica_mfu[replica_id - 1][stage_id - 1].put(time, 0)



from random import randint


class CollectivesInput:
    used_comm_ids = set()

    def __init__(
        self,
        num_workers: int,
        num_workers_per_node: int,
        collective_size: int,
        collective: str,
    ):
        self.num_workers = num_workers
        self.num_workers_per_node = num_workers_per_node
        self.collective_size = collective_size
        self.collective = collective
        self.comm_id = self._get_comm_id()

    @classmethod
    def _get_comm_id(cls):
        comm_id = randint(0, 65535)
        while comm_id in cls.used_comm_ids:
            comm_id = randint(0, 65535)
        cls.used_comm_ids.add(comm_id)
        return comm_id

    def is_valid(self, total_gpus_available: int, num_nodes: int):
        if self.num_workers == 1:
            return False

        if self.collective == "send_recv" and self.num_workers != 2:
            return False

        if (
            self.num_workers > self.num_workers_per_node
            and self.num_workers % self.num_workers_per_node != 0
        ):
            return False

        num_nodes_required = self.num_workers // self.num_workers_per_node

        if self.num_workers > total_gpus_available or num_nodes_required > num_nodes:
            return False

        return True



from vidur.entities import Request
from vidur.scheduler.replica_scheduler.mnemosyne_base_replica_scheduler import (
    BatchFormationTracker,
)
from vidur.scheduler.replica_scheduler.mnemosyne_lrs_replica_scheduler import (
    MnemosyneLRSReplicaScheduler,
)

MAX_SPACE_SHARE_FRAC = 0.5


class MnemosyneSTReplicaScheduler(MnemosyneLRSReplicaScheduler):

    def _get_request_next_num_q_tokens(
        self,
        time: float,
        request: Request,
        batch_formation_tracker: BatchFormationTracker,
    ) -> int:
        assert not request.completed

        if self._is_prefill_complete(request):
            return 1

        active_kvp_groups = self._get_active_kvp_group_ids(request)
        num_processed_tokens = self._get_num_processed_tokens(request)

        if num_processed_tokens < self._config.long_request_kv_cache_len_threshold:
            target_time = self._config.target_batch_time
        else:
            # avoid space sharing with another long request
            if any(
                any(
                    x > self._config.long_request_kv_cache_len_threshold
                    for x in batch_formation_tracker.batch_request_num_processed_tokens[
                        kvp_group_id
                    ]
                )
                for kvp_group_id in active_kvp_groups
            ):
                return 0

            slack_fraction = self._get_remaining_slack_fraction(time, request)
            slack_fraction = max(0.0, slack_fraction)
            slack_fraction = min(MAX_SPACE_SHARE_FRAC, slack_fraction)
            target_time = self._config.target_batch_time * (1 - slack_fraction)

        next_num_tokens = batch_formation_tracker.get_max_chunk_size_for_request(
            request,
            num_processed_tokens,
            active_kvp_groups,
            target_time,
        )

        next_num_tokens = max(0, next_num_tokens)

        return next_num_tokens



import time

import torch
from torch.profiler import record_function

from vidur.profiling.common.timer_stats_store import TimerStatsStore
from vidur.profiling.utils import ProfileMethod


class CudaTimer:
    def __init__(
        self,
        name,
        layer_id: int = 0,  # we don't care about layer id, it is just for compatibility with sarathi cudatimer
        aggregation_fn=sum,
        filter_str=None,
    ):
        if name:
            # beautify the names we get from vllm
            name = str(name).replace("OperationMetrics.", "")
            name = name.lower()
            self.name = f"vidur_{name}"
        else:
            self.name = None

        self.timer_stats_store = TimerStatsStore()
        self.disabled = (name is None) or self.timer_stats_store.disabled

        if self.disabled:
            return

        self.aggregation_fn = aggregation_fn
        self.filter_str = filter_str

        if self.timer_stats_store.profile_method == ProfileMethod.KINETO:
            self.profiler = torch.profiler.profile(
                activities=[torch.profiler.ProfilerActivity.CUDA],
                on_trace_ready=self.handle_trace,
            )
        else:
            self.profiler = None
        self.start_event = None
        self.end_event = None
        self.start_time = None
        self.end_time = None

    def __enter__(self):
        if self.disabled:
            return

        if self.timer_stats_store.profile_method == ProfileMethod.RECORD_FUNCTION:
            self.profiler_function_context = record_function(self.name)
            self.profiler_function_context.__enter__()
        elif self.timer_stats_store.profile_method == ProfileMethod.CUDA_EVENT:
            self.start_event = torch.cuda.Event(enable_timing=True)
            self.start_event.record()
        elif self.timer_stats_store.profile_method == ProfileMethod.KINETO:
            self.profiler.__enter__()
        elif self.timer_stats_store.profile_method == ProfileMethod.PERF_COUNTER:
            torch.cuda.synchronize()
            self.start_time = time.perf_counter()
        else:
            raise ValueError(
                f"Unknown profile method {self.timer_stats_store.profile_method}"
            )
        return self

    def handle_trace(self, trace):
        events = trace.events()

        if self.filter_str:
            events = [e for e in events if e.name.startswith(self.filter_str)]

        total_cuda_time = self.aggregation_fn([e.cuda_time_total for e in events])
        self.timer_stats_store.record_time(
            self.name, total_cuda_time * 1e-3
        )  # convert to ms

    def __exit__(self, *args):
        if self.disabled:
            return

        if self.timer_stats_store.profile_method == ProfileMethod.RECORD_FUNCTION:
            self.profiler_function_context.__exit__(*args)
        elif self.timer_stats_store.profile_method == ProfileMethod.CUDA_EVENT:
            self.end_event = torch.cuda.Event(enable_timing=True)
            self.end_event.record()
            self.timer_stats_store.record_time(
                self.name, [self.start_event, self.end_event]
            )
        elif self.timer_stats_store.profile_method == ProfileMethod.KINETO:
            self.profiler.__exit__(*args)
        elif self.timer_stats_store.profile_method == ProfileMethod.PERF_COUNTER:
            torch.cuda.synchronize()
            self.end_time = time.perf_counter()
            self.timer_stats_store.record_time(
                self.name, (self.end_time - self.start_time) * 1e3
            )  # convert to ms
        else:
            raise ValueError(
                f"Unknown profile method {self.timer_stats_store.profile_method}"
            )



import argparse
import glob
import os
import platform
import shlex
from subprocess import Popen

import pandas as pd
import ray

from vidur.config_optimizer.config_explorer.config import JobConfig, SimulationConfig
from vidur.config_optimizer.config_explorer.ray_utils import (
    CpuAssignmentManager,
    get_ip,
)
from vidur.logger import init_logger

logger = init_logger(__name__)


class CapacitySearch:
    def __init__(
        self,
        job_config: JobConfig,
        args: argparse.Namespace,
        cpu_core_assignment_manager: CpuAssignmentManager = None,
        cpu_core_id: int = None,
    ):
        self.node_ip = get_ip()
        self.cpu_core_id = None
        self.job_config = job_config
        self.args = args
        self.cpu_core_assignment_manager = cpu_core_assignment_manager
        self.cpu_core_id = cpu_core_id

    def release_cpu_core_id(self):
        if self.cpu_core_id is None:
            return

        ray.get(
            self.cpu_core_assignment_manager.release_cpu_core_id.remote(
                self.node_ip,
                self.cpu_core_id,
            )
        )

    def _generate_run_command(
        self,
        scheduler_config: SimulationConfig,
    ):
        cpu_affinity_command = ""
        if self.cpu_core_id is not None and platform.system() != "Darwin":
            cpu_affinity_command = f"taskset --cpu-list {self.cpu_core_id}"

        command = f"nice -n 1 {cpu_affinity_command} python -m vidur.main {scheduler_config.to_args()}"
        logger.debug(f"Running command: {command}")

        return command

    def _get_result_file(self, run_dir: str) -> str:
        scheduling_delay_file = glob.glob(
            f"{run_dir}/*/plots/request_scheduling_delay.csv"
        )
        if len(scheduling_delay_file) == 0:
            return

        return scheduling_delay_file[0]

    def _is_under_sla(
        self,
        result_file: str,
        simulator_config: SimulationConfig,
    ) -> tuple[bool, float]:
        scheduling_delay_df = pd.read_csv(result_file)
        scheduling_delay = scheduling_delay_df["request_scheduling_delay"].quantile(
            self.args.scheduling_delay_slo_quantile
        )
        is_under_scheduling_delay_sla = (
            scheduling_delay <= self.args.scheduling_delay_slo_value
        )

        logger.info(
            f"{simulator_config.to_human_readable_name()} - Scheduling delay (P{self.args.scheduling_delay_slo_quantile}): {scheduling_delay}",
        )
        return is_under_scheduling_delay_sla, scheduling_delay

    def is_under_sla(self, qps: float) -> tuple[bool, float]:
        simulator_config = SimulationConfig(
            output_dir=self.args.output_dir,
            cache_dir=self.args.cache_dir,
            qps=qps,
            time_limit=self.args.time_limit,
            job_config=self.job_config,
        )
        run_dir = simulator_config.get_run_dir()
        os.makedirs(run_dir, exist_ok=True)

        cached_result_file = self._get_result_file(run_dir)
        if cached_result_file:
            return self._is_under_sla(cached_result_file, simulator_config)

        command = self._generate_run_command(simulator_config)

        output_file = open(f"{run_dir}/output.log", "w")

        # write command to a file
        output_file.write(f"Running command: {command}\n")

        try:
            args = shlex.split(command)
            p = Popen(args, stdout=output_file, stderr=output_file)
            p.wait()

            result_file = self._get_result_file(run_dir)
            assert (
                result_file is not None
            ), f"Result file not found for {simulator_config.to_human_readable_name()}"
            return self._is_under_sla(result_file, simulator_config)
        except Exception as e:
            logger.error(
                f"Error running: {self.job_config.get_human_readable_name()}, failed with error: {e}",
            )
            return False, None

    def search(self):
        """
        Perform binary search to find the maximum QPS under the SLO
        """
        logger.info(
            f"Starting search for {self.job_config.get_human_readable_name()}",
        )

        left = 0
        right = self.job_config.start_qps * 2
        qps = 0
        max_qps_under_sla = None
        min_qps_over_sla = 2**32

        for _ in range(self.args.max_iterations):
            # stopping condition - we have reached the minimum granularity
            if abs(left - right) < self.args.min_search_granularity * qps / 100:
                break

            qps = (left + right) / 2

            is_under_sla, scheduling_delay = self.is_under_sla(qps)

            if scheduling_delay is None:
                break

            if is_under_sla:
                max_qps_under_sla = qps

                if scheduling_delay < self.args.scheduling_delay_slo_value / 8:
                    # if the scheduling delay is very low, we can increase the QPS more aggressively
                    right = min(right * 4, min_qps_over_sla)
                elif scheduling_delay < self.args.scheduling_delay_slo_value / 4:
                    right = min(right * 2, min_qps_over_sla)
                elif qps > 0.8 * right:
                    right = min(right * 2, min_qps_over_sla)

                left = qps
            else:
                if scheduling_delay > 500:
                    right = qps / 2
                elif scheduling_delay > 1000:
                    right = qps / 4
                else:
                    right = qps

                min_qps_over_sla = min(min_qps_over_sla, qps)

        logger.info(
            f"Max QPS under SLO for {self.job_config.get_human_readable_name()}: {max_qps_under_sla}",
        )

        self.release_cpu_core_id()

        return {
            **self.job_config.to_config_dict(),
            "max_qps_under_sla": max_qps_under_sla,
        }



import heapq
import json
from typing import List

from vidur.config import SimulationConfig
from vidur.entities import Cluster
from vidur.events import BaseEvent, RequestArrivalEvent
from vidur.logger import init_logger
from vidur.metrics import MetricsStore
from vidur.request_generator import RequestGeneratorRegistry
from vidur.scheduler import BaseGlobalScheduler, GlobalSchedulerRegistry

logger = init_logger(__name__)


class Simulator:
    def __init__(self, config: SimulationConfig) -> None:
        self._config: SimulationConfig = config

        self._time = 0
        self._terminate = False
        self._time_limit = self._config.time_limit
        if not self._time_limit:
            self._time_limit = float("inf")

        self._event_queue = []

        self._event_trace = []
        self._event_chrome_trace = []

        self._cluster = Cluster(
            self._config.cluster_config,
            self._config.metrics_config,
            self._config.request_generator_config,
        )
        self._metric_store = MetricsStore(self._config)
        self._request_generator = RequestGeneratorRegistry.get(
            self._config.request_generator_config.get_type(),
            self._config.request_generator_config,
        )
        self._scheduler = GlobalSchedulerRegistry.get(
            self._config.cluster_config.global_scheduler_config.get_type(),
            self._config,
            self._cluster.replicas,
        )

        self._init_event_queue()

    @property
    def scheduler(self) -> BaseGlobalScheduler:
        return self._scheduler

    @property
    def metric_store(self) -> MetricsStore:
        return self._metric_store

    def run(self) -> None:
        logger.info(
            f"Starting simulation with cluster: {self._cluster} and {len(self._event_queue)} requests"
        )

        while self._event_queue and not self._terminate:
            _, event = heapq.heappop(self._event_queue)
            self._set_time(event._time)
            new_events = event.handle_event(self._scheduler, self._metric_store)
            self._add_events(new_events)

            if self._config.metrics_config.write_json_trace:
                self._event_trace.append(event.to_dict())

            if self._config.metrics_config.enable_chrome_trace:
                trace_events = event.to_chrome_trace_events()
                if trace_events:
                    self._event_chrome_trace.extend(trace_events)

        assert (
            self._scheduler.is_empty() or self._terminate
        ), f"Simulation ended with pending requests {self._scheduler.get_pending_requests()}"

        logger.info(f"Simulation ended at: {self._time}s")

        self._write_output()

    def _write_output(self) -> None:
        logger.info("Writing output")

        self._metric_store.plot()
        logger.info("Metrics written")

        if self._config.metrics_config.write_json_trace:
            self._write_event_trace()
            self._scheduler.write_batching_history()
            logger.info("Json event trace written")

        if self._config.metrics_config.enable_chrome_trace:
            self._write_chrome_trace()
            logger.info("Chrome event trace written")

    def _add_event(self, event: BaseEvent) -> None:
        heapq.heappush(self._event_queue, (event._priority_number, event))

    def _add_events(self, events: List[BaseEvent]) -> None:
        for event in events:
            self._add_event(event)

    def _init_event_queue(self) -> None:
        requests = self._request_generator.generate()

        for request in requests:
            self._add_event(RequestArrivalEvent(request.arrived_at, request))

    def _set_time(self, time: float) -> None:
        self._time = time
        if self._time > self._time_limit:
            logger.info(
                f"Time limit reached: {self._time_limit}s terminating the simulation."
            )
            self._terminate = True

    def _write_event_trace(self) -> None:
        trace_file = f"{self._config.metrics_config.output_dir}/event_trace.json"
        with open(trace_file, "w") as f:
            json.dump(self._event_trace, f)

    def _write_chrome_trace(self) -> None:
        trace_file = f"{self._config.metrics_config.output_dir}/chrome_trace.json"

        chrome_trace = {"traceEvents": self._event_chrome_trace}

        with open(trace_file, "w") as f:
            json.dump(chrome_trace, f)



import os

import sarathi.metrics.cuda_timer
import torch

from vidur.profiling.common.cuda_timer import CudaTimer

# monkey patching the CudaTimer class to use the sarathi implementation
sarathi.metrics.cuda_timer.CudaTimer = CudaTimer

from sarathi.model_executor.weight_utils import initialize_dummy_weights

from vidur.profiling.common.model_config import ModelConfig
from vidur.profiling.common.timer_stats_store import TimerStatsStore
from vidur.profiling.mlp.mlp_impl import GPTModel
from vidur.profiling.utils import ProfileMethod
from vidur.profiling.utils.record_function_tracer import RecordFunctionTracer

WARMUP_STEPS = 2
ACTIVE_STEPS = 20


class MlpWrapper:
    def __init__(
        self,
        model_config: ModelConfig,
        num_tensor_parallel_workers: int,
        profile_method: str,
        rank: int,
        output_dir: str,
    ):
        super().__init__()

        self.timer_stats_store = TimerStatsStore(profile_method=profile_method)

        self.model_config = model_config
        self.num_tensor_parallel_workers = num_tensor_parallel_workers
        self.profile_method = profile_method
        self.rank = rank
        self.output_dir = output_dir
        os.makedirs(f"{self.output_dir}/profiler_traces/", exist_ok=True)

        self.model = GPTModel(
            model_config,
            num_tensor_parallel_workers,
            (
                ACTIVE_STEPS
                if self.profile_method == ProfileMethod.RECORD_FUNCTION.value
                else 1
            ),
        )
        initialize_dummy_weights(self.model)
        self.model = self.model.to(dtype=torch.float16).cuda().eval()

    @torch.inference_mode()
    def profile(self, num_tokens: int):
        vocab_range = self.model_config.vocab_size // self.num_tensor_parallel_workers
        input_ids = torch.randint(
            low=0,
            high=vocab_range,
            size=(num_tokens,),
            device="cuda",
            dtype=torch.long,
        )
        positions = torch.arange(num_tokens, device="cuda", dtype=torch.long)

        if self.profile_method == ProfileMethod.RECORD_FUNCTION.value:
            # Run the model once without capturing the graph.
            # This is to make sure that the captured graph does not include the
            # kernel launches for initial benchmarking (e.g., Triton autotune).
            self.model(
                input_ids,
                positions,
            )
            torch.cuda.synchronize()

            self.timer_stats_store.clear_stats()

            record_function_tracer = RecordFunctionTracer(self.output_dir)

            with record_function_tracer:
                self.model(
                    input_ids,
                    positions,
                )

            time_stats = record_function_tracer.get_operation_time_stats()
        else:
            for _ in range(WARMUP_STEPS):
                self.model(
                    input_ids,
                    positions,
                )

            torch.cuda.synchronize()

            self.timer_stats_store.clear_stats()

            for _ in range(ACTIVE_STEPS):
                self.model(
                    input_ids,
                    positions,
                )

            torch.cuda.synchronize()

            time_stats = self.timer_stats_store.get_stats()

        stats = {
            "time_stats": time_stats,
            "n_head": self.model_config.num_q_heads,
            "n_kv_head": self.model_config.num_kv_heads,
            "n_embd": self.model_config.embedding_dim,
            "n_expanded_embd": self.model_config.mlp_hidden_dim,
            "vocab_size": self.model_config.vocab_size,
            "use_gated_mlp": self.model_config.use_gated_mlp,
            "num_tokens": num_tokens,
            "num_tensor_parallel_workers": self.num_tensor_parallel_workers,
        }
        self.timer_stats_store.clear_stats()

        return stats



from abc import ABC, abstractmethod
from typing import List, Union

from vidur.config import (
    BaseReplicaSchedulerConfig,
    BaseRequestGeneratorConfig,
    ReplicaConfig,
)
from vidur.entities import (
    Batch,
    BatchStage,
    KVParallelBatch,
    KVParallelBatchStage,
    Replica,
    Request,
)
from vidur.execution_time_predictor import BaseExecutionTimePredictor
from vidur.logger import init_logger
from vidur.scheduler.replica_stage_scheduler import ReplicaStageScheduler
from vidur.scheduler.utils.memory_planner import MemoryPlanner

logger = init_logger(__name__)


class BaseReplicaScheduler(ABC):
    def __init__(
        self,
        replica_config: ReplicaConfig,
        replica_scheduler_config: BaseReplicaSchedulerConfig,
        request_generator_config: BaseRequestGeneratorConfig,
        replica: Replica,
        num_stages: int,
        execution_time_predictor: BaseExecutionTimePredictor,
    ) -> None:
        self._config = replica_scheduler_config
        self._replica_config = replica_config
        self._request_generator_config = request_generator_config
        self._replica_id = replica.id
        self._num_stages = num_stages
        self._execution_time_predictor = execution_time_predictor

        self._max_blocks_per_sequence = (
            self._request_generator_config.max_tokens // self._replica_config.block_size
        )

        memory_planner = MemoryPlanner(self._replica_config, replica)

        if not self._config.num_blocks:
            self._config.num_blocks = (
                self._max_blocks_per_sequence * memory_planner.get_max_request_slots()
            )
        self._max_batch_size = min(
            memory_planner.get_max_batch_size(),
            self._config.batch_size_cap,
        )

        logger.debug(
            f"Obtained max batch size of {self._max_batch_size} for replica {self._replica_id}"
        )

        self._request_queue = []
        self._num_allocated_blocks = 0
        self._allocation_map = {}

        self._replica_stage_schedulers = {
            stage_id: ReplicaStageScheduler(
                replica.id,
                stage_id,
                stage_id == num_stages - 1,
                execution_time_predictor,
            )
            for stage_id in range(num_stages)
        }

        self._preempted_requests: List[Request] = []

        self._num_running_batches = 0
        self._num_running_stages = 0

    @property
    def num_pending_requests(self) -> int:
        return len(self._request_queue)

    @property
    def replica_id(self) -> int:
        return self._replica_id

    @property
    def num_allocated_blocks(self) -> int:
        return self._num_allocated_blocks

    @property
    def memory_usage_percent(self) -> int:
        return (self._num_allocated_blocks * 100) / self._config.num_blocks

    @property
    def is_sequence_pipeline_parallel_enabled(self) -> bool:
        return self._replica_config.enable_sequence_pipeline_parallel

    def is_empty(self) -> bool:
        return (
            self.num_pending_requests == 0
            and len(self._allocation_map) == 0
            and all(
                stage_scheduler.is_empty()
                for stage_scheduler in self._replica_stage_schedulers.values()
            )
        )

    def _get_request_next_num_q_tokens(self, request: Request) -> int:
        assert not request.completed

        if request.is_prefill_complete:
            return 1

        return request.num_prefill_tokens

    def _get_request_next_num_kv_tokens(
        self, request: Request, num_q_tokens: int
    ) -> int:
        return num_q_tokens + request.num_processed_tokens

    def add_request(self, request: Request) -> None:
        self._request_queue.append(request)

    def get_pending_requests(self) -> List[Request]:
        return self._request_queue

    def get_replica_stage_scheduler(self, stage_id: int):
        return self._replica_stage_schedulers[stage_id]

    def can_allocate(self, num_blocks: int) -> bool:
        return self._config.num_blocks - self._num_allocated_blocks >= num_blocks

    def allocate(self, request_id: int, num_blocks: int) -> None:
        self._num_allocated_blocks += num_blocks
        if request_id not in self._allocation_map:
            self._allocation_map[request_id] = num_blocks
        else:
            self._allocation_map[request_id] += num_blocks

        assert self._num_allocated_blocks <= self._config.num_blocks

    def free(self, *request_ids: List[int]) -> None:
        for request_id in request_ids:
            num_blocks = self._allocation_map.pop(request_id)
            self._num_allocated_blocks -= num_blocks

        assert self._num_allocated_blocks >= 0

    def free_batch(self, batch: Union[Batch, KVParallelBatch]) -> None:
        self.free(*batch.request_ids)

    def on_batch_end(self, batch: Union[Batch, KVParallelBatch]) -> None:
        self._num_running_batches -= 1
        assert self._num_running_batches >= 0

    def on_stage_end(
        self, stage_id: int, batch_stage: Union[BatchStage, KVParallelBatchStage]
    ) -> None:
        if not stage_id == 0:
            return

        self._num_running_stages -= 1

    @abstractmethod
    def _get_next_batch(self, time: float) -> Union[Batch, KVParallelBatch]:
        pass

    def on_schedule(self, time: float) -> Union[Batch, KVParallelBatch]:
        if not (
            self._num_running_batches < self._num_stages
            and self._num_running_stages == 0
        ):
            return None

        batch = self._get_next_batch(time)
        if not batch:
            return None

        self._num_running_batches += 1
        self._num_running_stages += 1
        return batch



import binascii
import enum
from itertools import product
from math import floor
from typing import List

import torch

from vidur.profiling.attention.attention_input import AttentionInput
from vidur.profiling.collectives.collectives_input import CollectivesInput


class ProfileMethod(enum.Enum):
    CUDA_EVENT = "cuda_event"
    KINETO = "kineto"
    PERF_COUNTER = "perf_counter"
    RECORD_FUNCTION = "record_function"


def get_num_tokens_to_profile(
    max_num_tokens: int,
):
    NUM_TOKENS_SPACE = (
        list([1, 2, 4])
        + list(range(8, 1024, 8))
        + list(range(1024, 2 * 1024 + 1, 16))
        + list(range(2 * 1024, 4 * 1024 + 1, 32))
        + list(range(4 * 1024, 8 * 1024 + 1, 64))
        + list(range(8 * 1024, 16 * 1024 + 1, 128))
        + list(range(16 * 1024, 32 * 1024 + 1, 256))
        + list(range(32 * 1024, 64 * 1024 + 1, 512))
        + list(range(64 * 1024, 128 * 1024 + 1, 1024))
    )
    num_tokens_to_profile = []
    for num_tokens in NUM_TOKENS_SPACE:
        if num_tokens <= max_num_tokens:
            num_tokens_to_profile.append(num_tokens)
        else:
            break
    num_tokens_to_profile.sort(reverse=True)

    return num_tokens_to_profile


def get_attention_batch_sizes_to_profile(min_batch_size: int, max_batch_size: int):
    BATCH_SIZE_SPACE = list(range(1, 128 + 1, 1)) + list(range(128, 1024 + 1, 8))
    return list(
        filter(
            lambda x: (x >= min_batch_size and x <= max_batch_size), BATCH_SIZE_SPACE
        )
    )


def get_attention_prefill_chunk_sizes_to_profile(max_seq_len: int, max_chunk_size: int):
    PREFILL_CHUNK_SIZE_SPACE = (
        list(range(32, 128 + 1, 32))
        + list(range(256, 1024 + 1, 128))
        + list(range(1024, 8 * 1024 + 1, 512))
    )  # 32 - 8k
    prefill_chunk_sizes_to_profile = []
    for prefill_chunk_size in PREFILL_CHUNK_SIZE_SPACE:
        if prefill_chunk_size <= max_seq_len and prefill_chunk_size <= max_chunk_size:
            prefill_chunk_sizes_to_profile.append(prefill_chunk_size)
        else:
            break
    return prefill_chunk_sizes_to_profile


def get_seq_lengths_to_profile(max_seq_len: int):
    SEQ_LENGTH_SIZE_SPACE = (
        list(range(0, 1024 + 1, 32))
        + list(range(1024, 4 * 1024 + 1, 64))
        + list(range(4 * 1024, 64 * 1024 + 1, 256))
        + list(range(64 * 1024, 512 * 1024 + 1, 1024))
        + list(range(512 * 1024, 4096 * 1024 + 1, 4096))
        + list(range(4096 * 1024, 1100 * 1024 + 1, 16 * 1024))
    )
    seq_lengths_to_profile = []
    for seq_length in SEQ_LENGTH_SIZE_SPACE:
        if seq_length < max_seq_len:
            seq_lengths_to_profile.append(seq_length)
        else:
            break
    return seq_lengths_to_profile


def get_attention_input_combinations(
    max_seq_len: int,
    min_batch_size: int,
    max_batch_size: int,
    profile_only_prefill: bool,
    profile_only_decode: bool,
    max_chunk_size: int,
):
    input_combinations = []
    # Chunked Prefills
    prefill_chunk_sizes_to_profile = get_attention_prefill_chunk_sizes_to_profile(
        max_seq_len, max_chunk_size
    )
    for prefill_chunk_size in prefill_chunk_sizes_to_profile:
        num_partitions = max_seq_len // prefill_chunk_size
        kv_cache_sizes_to_profile = [
            partition_index * prefill_chunk_size
            for partition_index in range(num_partitions)
        ]
        input_combinations.extend(
            product([prefill_chunk_size], kv_cache_sizes_to_profile, [1], [True])
        )
    # Full prefills
    if max_seq_len <= max_chunk_size:
        prefill_lengths_to_profile = get_seq_lengths_to_profile(max_seq_len)
        input_combinations.extend(product(prefill_lengths_to_profile, [0], [1], [True]))
    # Decodes
    kv_cache_sizes_to_profile = get_seq_lengths_to_profile(max_seq_len)
    batch_sizes_to_profile = get_attention_batch_sizes_to_profile(
        min_batch_size, max_batch_size
    )
    input_combinations.extend(
        product([0], kv_cache_sizes_to_profile, batch_sizes_to_profile, [False])
    )

    valid_input_combinations = []
    for input_combination in input_combinations:
        prefill_chunk_size, kv_cache_size, batch_size, is_prefill = input_combination

        if is_prefill and profile_only_decode:
            continue

        if not is_prefill and profile_only_prefill:
            continue

        attention_input = AttentionInput(
            prefill_chunk_size,
            kv_cache_size,
            batch_size,
            is_prefill,
        )

        if attention_input.is_valid(max_seq_len):
            valid_input_combinations.append(attention_input)
    return valid_input_combinations


"""
    For a given model and parallel config, get the maximum number of blocks that can be allocated.
    This doesn't take into account the weights and activations.
"""


def get_max_num_blocks(
    model_config: "ModelConfig",
    parallel_config: "ParallelConfig",
    block_size: int,
    dtype: torch.dtype,
    gpu_memory_utilization: float = 0.9,
    max_pipeline_parallel_size: int = 8,
):
    element_size = torch.randn(1, dtype=dtype).element_size()
    block_memory_size = (
        2
        * block_size
        * model_config.get_num_kv_heads(parallel_config)
        * model_config.get_head_size()
        * element_size
    )
    assert model_config.num_layers % max_pipeline_parallel_size == 0
    block_memory_total = block_memory_size * (
        model_config.num_layers // max_pipeline_parallel_size
    )
    return floor(
        (torch.cuda.mem_get_info()[1] * gpu_memory_utilization) / (block_memory_total)
    )


def get_collectives_sizes_to_profile(max_collective_size: int):
    COLLECTIVE_SIZE_SPACE = (
        list(range(1024, 512 * 1024 + 1, 4 * 1024))
        + list(range(512 * 1024, 8 * 1024 * 1024 + 1, 16 * 1024))
        + list(range(8 * 1024 * 1024, 64 * 1024 * 1024 + 1, 64 * 1024))
        + list(range(64 * 1024 * 1024 + 1, 512 * 1024 * 1024 + 1, 265 * 1024))
    )
    collectives_size_to_profile = []
    for collectives_size in COLLECTIVE_SIZE_SPACE:
        if collectives_size <= max_collective_size:
            collectives_size_to_profile.append(collectives_size)
        else:
            break
    return collectives_size_to_profile


def get_collectives_inputs(
    num_nodes: int,
    num_workers_per_node_combinations: List[int],
    max_collective_size: int,
    collective: str,
    total_gpus_available: int,
):
    num_workers = []

    for num_workers_per_node in num_workers_per_node_combinations:
        for _num_nodes in range(1, num_nodes + 1):
            num_workers.append(num_workers_per_node * _num_nodes)

    num_workers = list(set(num_workers))
    collectives_sizes = get_collectives_sizes_to_profile(max_collective_size)

    collectives_inputs = []

    for num_workers, num_workers_per_node, collective_size in product(
        num_workers, num_workers_per_node_combinations, collectives_sizes
    ):
        collectives_input = CollectivesInput(
            num_workers, num_workers_per_node, collective_size, collective
        )
        if not collectives_input.is_valid(total_gpus_available, num_nodes):
            continue

        collectives_inputs.append(collectives_input)

    return collectives_inputs


def hex_to_binary(hex_identifier):
    return binascii.unhexlify(hex_identifier)



from typing import List, Tuple

from vidur.entities import Request
from vidur.scheduler.global_scheduler.base_global_scheduler import BaseGlobalScheduler


class RoundRobinGlobalScheduler(BaseGlobalScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._request_counter = 0

    def schedule(self) -> List[Tuple[int, Request]]:
        self.sort_requests()

        request_mapping = []
        while self._request_queue:
            request = self._request_queue.pop(0)
            replica_id = self._request_counter % self._num_replicas
            self._request_counter += 1
            request_mapping.append((replica_id, request))

        return request_mapping



"""
Singleton metaclass as described in
https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
"""


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from vidur.scheduler.global_scheduler.base_global_scheduler import BaseGlobalScheduler
from vidur.scheduler.global_scheduler.global_scheduler_registry import (
    GlobalSchedulerRegistry,
)

__all__ = [BaseGlobalScheduler, GlobalSchedulerRegistry]



from typing import List

from vidur.entities import Batch
from vidur.events import BaseEvent
from vidur.logger import init_logger
from vidur.metrics import MetricsStore
from vidur.scheduler import BaseGlobalScheduler
from vidur.types import EventType

logger = init_logger(__name__)


class BatchEndEvent(BaseEvent):
    def __init__(self, time: float, replica_id: int, batch: Batch):
        super().__init__(time, EventType.BATCH_END)

        self._replica_id = replica_id
        self._batch = batch

    def handle_event(
        self, scheduler: BaseGlobalScheduler, metrics_store: MetricsStore
    ) -> List[BaseEvent]:
        from vidur.events.replica_schedule_event import ReplicaScheduleEvent

        self._batch.on_batch_end(self.time)
        replica_scheduler = scheduler.get_replica_scheduler(self._replica_id)
        replica_scheduler.on_batch_end(self._batch)

        memory_usage_percent = replica_scheduler.memory_usage_percent
        metrics_store.on_batch_end(
            self.time, self._batch, self._replica_id, memory_usage_percent
        )

        return [ReplicaScheduleEvent(self.time, self._replica_id)]

    def to_dict(self):
        return {
            "time": self.time,
            "event_type": self.event_type,
            "batch_id": self._batch.id,
        }



from vidur.request_generator.base_request_interval_generator import (
    BaseRequestIntervalGenerator,
)


class StaticRequestIntervalGenerator(BaseRequestIntervalGenerator):

    def get_next_inter_request_time(self) -> float:
        return 0



import json

import wandb

from vidur.logger import init_logger

logger = init_logger(__name__)


class SeriesAverageMeter:
    def __init__(
        self,
        x_name: str,
        y_name: str,
        use_weighted_mean: bool = True,
        save_table_to_wandb: bool = True,
    ) -> None:
        # column names of x, y datatpoints for data collection
        self._x_name = x_name
        self._y_name = y_name
        self._use_weighted_mean = use_weighted_mean
        self._save_table_to_wandb = save_table_to_wandb

        self._denom_sum = 0
        self._numer_sum = 0
        self._min_y = float("inf")
        self._max_y = float("-inf")
        # most recently collected y datapoint for incremental updates
        # to aid incremental updates to y datapoints
        self._last_data_y = None
        self._last_data_x = None

    def _update_simple_mean(self, data_y: float) -> None:
        self._denom_sum += 1
        self._numer_sum += data_y

    def _update_weighted_mean(self, data_x: float) -> None:
        if not self._last_data_x:
            return

        # compute the segment length
        x_diff = data_x - self._last_data_x
        # Add the weighted value multiplied by the time difference to the total weighted sum
        self._numer_sum += self._last_data_y * x_diff
        # Add the time difference to the total time
        self._denom_sum += x_diff

    # add a new x, y datapoint
    def put(self, data_x: float, data_y: float) -> None:
        if self._use_weighted_mean:
            self._update_weighted_mean(data_x)
        else:
            self._update_simple_mean(data_y)

        self._last_data_y = data_y
        self._last_data_x = data_x
        self._min_y = min(self._min_y, data_y)
        self._max_y = max(self._max_y, data_y)

    # get most recently collected y datapoint
    def _peek_y(self):
        return self._last_data_y

    # add a new x, y datapoint as an incremental (delta) update to
    # recently collected y datapoint
    def put_delta(self, data_x: float, data_y_delta: float) -> None:
        last_data_y = self._peek_y()
        data_y = last_data_y + data_y_delta
        self.put(data_x, data_y)

    def print_stats(self, name: str, path: str) -> None:
        if self._denom_sum == 0:
            return

        weighted_mean = self._numer_sum / self._denom_sum

        logger.debug(
            f"{name}: {self._y_name} stats:"
            f" min: {self._min_y},"
            f" max: {self._max_y},"
            f" weighted_mean: {weighted_mean}"
        )

        stats_dict = {
            f"{name}_min": self._min_y,
            f"{name}_max": self._max_y,
            f"{name}_weighted_mean": weighted_mean,
        }

        with open(f"{path}/{name}.json", "w") as f:
            json.dump(stats_dict, f)

        if wandb.run:
            wandb.log(
                {
                    f"{name}_min": self._min_y,
                    f"{name}_max": self._max_y,
                    f"{name}_weighted_mean": weighted_mean,
                },
                step=0,
            )



import numpy as np

from vidur.profiling.utils import ProfileMethod
from vidur.profiling.utils.singleton import Singleton


class TimerStatsStore(metaclass=Singleton):
    def __init__(self, profile_method: str, disabled: bool = False):
        self.disabled = disabled
        self.profile_method = ProfileMethod[profile_method.upper()]
        self.TIMING_STATS = {}

    def record_time(self, name: str, time):
        name = name.replace("vidur_", "")
        if name not in self.TIMING_STATS:
            self.TIMING_STATS[name] = []

        self.TIMING_STATS[name].append(time)

    def clear_stats(self):
        self.TIMING_STATS = {}

    def get_stats(self):
        stats = {}
        for name, times in self.TIMING_STATS.items():
            times = [
                (time if isinstance(time, float) else time[0].elapsed_time(time[1]))
                for time in times
            ]

            stats[name] = {
                "min": np.min(times),
                "max": np.max(times),
                "mean": np.mean(times),
                "median": np.median(times),
                "std": np.std(times),
            }

        return stats



from sklearn.linear_model import LinearRegression
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures

from vidur.config import LinearRegressionExecutionTimePredictorConfig, ReplicaConfig
from vidur.execution_time_predictor.sklearn_execution_time_predictor import (
    SklearnExecutionTimePredictor,
)


class LinearRegressionExecutionTimePredictor(SklearnExecutionTimePredictor):
    def __init__(
        self,
        predictor_config: LinearRegressionExecutionTimePredictorConfig,
        replica_config: ReplicaConfig,
    ) -> None:
        # will trigger model training
        super().__init__(
            predictor_config=predictor_config,
            replica_config=replica_config,
        )

    def _get_grid_search_params(self):
        return {
            "polynomialfeatures__degree": self._config.polynomial_degree,
            "polynomialfeatures__include_bias": self._config.polynomial_include_bias,
            "polynomialfeatures__interaction_only": self._config.polynomial_interaction_only,
            "linearregression__fit_intercept": self._config.fit_intercept,
        }

    def _get_estimator(self):
        return make_pipeline(PolynomialFeatures(), LinearRegression())



from sklearn.ensemble import RandomForestRegressor

from vidur.config import RandomForrestExecutionTimePredictorConfig, ReplicaConfig
from vidur.entities import Batch
from vidur.execution_time_predictor.sklearn_execution_time_predictor import (
    SklearnExecutionTimePredictor,
)


class RandomForrestExecutionTimePredictor(SklearnExecutionTimePredictor):
    def __init__(
        self,
        predictor_config: RandomForrestExecutionTimePredictorConfig,
        replica_config: ReplicaConfig,
    ) -> None:
        # will trigger model training
        super().__init__(
            predictor_config=predictor_config,
            replica_config=replica_config,
        )

    def _get_grid_search_params(self):
        return {
            "n_estimators": self._config.num_estimators,
            "max_depth": self._config.max_depth,
            "min_samples_split": self._config.min_samples_split,
        }

    def _get_estimator(self):
        return RandomForestRegressor()

    def _get_kv_parallel_communication_time(self, batch: Batch) -> float:
        return super()._get_kv_parallel_communication_time(batch)



from vidur.config_optimizer.analyzer.bottleneck_cases import *


class BottleneckAnalyzer:
    """
    Why I can't get more capacity out of my replica?
    C1. Violates TFFT SLO
        C1.1 Limited by  max batch size - Batch size saturated - mean memory usage is low.
        C1.2 Limited by memory availability - Memory usage is high, batch size is lower than limit.
        C1.3 Limited by prefill throughput - None of the above applies (NEED BETTER STRAT).
            C1.3.1 Low efficiency due to pipeline bubbles - high idle time and pipeline parallelism.
            X C1.3.2 Low efficiency due to network latency - high TP comm time.
            X C1.3.3 Low efficiency due to chunking overhead - lower prefill MFU and small chunk size.
            X C1.3.4 Slow prefill due to low parallelism - high prefill MFU and low TP degree.
    C2. Violates TBT SLO
        C2.1 High latency model execution - P25/P50 TBT is high.
            C2.1.1 Stalls due to pipeline bubbles - high idle time and pipeline parallelism.
            X C2.1.2 High latency due to network latency - lower MFU and high TP degree.
            X C2.1.4 Slow prefill due to low parallelism - high MFU and low TP degree.
            X C2.1.3 High latency due to chunking overhead - high chunk size.
        C2.2 High tail latency - lower percentiles are okay but tail latency is violated.
            C2.2.1 High tail latency due to scheduling problems - non-sarathi scheduler.
            C2.2.2 High tail latency due to large chunk size - sarathi scheduler and high chunk size.

    Classes marked with X are not implemented yet, and require more metrics to be implemented.
    """

    def __init__(
        self, ttft_slo_percentile, ttft_slo_value, tbt_slo_percentile, tbt_slo_value
    ):
        self.ttft_slo_percentile = ttft_slo_percentile
        self.ttft_slo_value = ttft_slo_value
        self.tbt_slo_percentile = tbt_slo_percentile
        self.tbt_slo_value = tbt_slo_value

    def analyze(self, config_row):
        """
        Analyze the config row and return the bottleneck.
        """
        params = (
            config_row,
            self.ttft_slo_percentile,
            self.ttft_slo_value,
            self.tbt_slo_percentile,
            self.tbt_slo_value,
        )

        # put child classes before parent classes
        match_seq = [
            TFFTViolationLowMaxBatchSizeCase,
            TFFTViolationLowMemoryCase,
            TFFTViolationLowPrefillThroughputPipelineBubblesCase,
            TFFTViolationLowPrefillThroughputCase,
            TBTViolationHighModelExecutionLatencyPipelineBubblesCase,
            TBTViolationHighModelExecutionLatencyCase,
            TBTViolationHighTailLatencySchedulerCase,
            TBTViolationHighTailLatencySarathiChunkSizeCase,
        ]

        # find and return the first match
        for case in match_seq:
            case_inst = case(*params)
            if case_inst.is_match():
                return case_inst.get_message()

        return "Could not interpret the bottleneck."



from math import ceil

from vidur.entities import Batch, BatchStage, Request
from vidur.scheduler.replica_scheduler.base_replica_scheduler import (
    BaseReplicaScheduler,
)


class SarathiReplicaScheduler(BaseReplicaScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # For vLLM and its derivatives, we only need to set a loose max batch size
        # Memory requirements are handled explicitly by the scheduler
        self._max_micro_batch_size = self._config.batch_size_cap // self._num_stages
        self._watermark_blocks = int(
            self._config.watermark_blocks_fraction * self._config.num_blocks
        )

    def _can_allocate_request(self, request: Request) -> bool:
        if request.id not in self._allocation_map:
            # new request
            num_required_blocks = ceil(
                request.num_prefill_tokens / self._replica_config.block_size
            )
            return (
                self._config.num_blocks
                - self._num_allocated_blocks
                - num_required_blocks
                >= self._watermark_blocks
            )

        # vllm requires at least one block to be available
        return self._config.num_blocks - self._num_allocated_blocks >= 1

    def _allocate_request(self, request: Request) -> None:
        if request.id not in self._allocation_map:
            # new request
            num_required_blocks = ceil(
                request.num_prefill_tokens / self._replica_config.block_size
            )
            self.allocate(request.id, num_required_blocks)
            return

        num_tokens_reserved = (
            self._allocation_map[request.id] * self._replica_config.block_size
        )
        num_tokens_required = max(0, request.num_processed_tokens - num_tokens_reserved)

        assert (
            num_tokens_required == 0 or num_tokens_required == 1
        ), f"num_tokens_required: {num_tokens_required}"

        if num_tokens_required == 0:
            return

        self.allocate(request.id, 1)

    def on_batch_end(self, batch: Batch) -> None:
        super().on_batch_end(batch)

        for request in batch.requests:
            if request.completed:
                self.free(request.id)
                continue

            if not self.is_sequence_pipeline_parallel_enabled:
                self._preempted_requests.append(request)
            else:
                if request.is_prefill_complete:
                    self._preempted_requests.append(request)

    def on_stage_end(self, stage_id: int, batch_stage: BatchStage) -> None:
        super().on_stage_end(stage_id, batch_stage)

        if not stage_id == 0:
            return

        if not self.is_sequence_pipeline_parallel_enabled:
            return

        for request in batch_stage.requests:
            assert not request.completed

            if request.is_prefill_stage_complete:
                continue
            self._preempted_requests.append(request)

    def _is_prefill_complete(self, request: Request) -> bool:
        if self.is_sequence_pipeline_parallel_enabled:
            return request.is_prefill_stage_complete
        else:
            return request.is_prefill_complete

    def _get_num_processed_prefill_tokens(self, request: Request) -> int:
        if self.is_sequence_pipeline_parallel_enabled:
            assert not request.is_prefill_stage_complete
            return request.num_stage_processed_tokens
        else:
            assert not request.is_prefill_complete
            return request.num_processed_tokens

    def _get_request_next_num_q_tokens(
        self, request: Request, num_batch_tokens: int
    ) -> int:
        assert not request.completed

        if self._is_prefill_complete(request):
            return 1

        next_num_tokens = min(
            request.num_prefill_tokens
            - self._get_num_processed_prefill_tokens(request),
            self._config.chunk_size - num_batch_tokens,
        )

        next_num_tokens = max(0, next_num_tokens)

        return next_num_tokens

    def _get_next_batch(self, time: float) -> Batch:
        requests = []
        num_q_tokens = []
        num_kv_tokens = []
        skipped_requests = []
        running_prefills = []
        total_num_q_tokens = 0

        if self._num_running_stages != 0:
            return

        # preempted requests could contain multiple requests which have
        # partial prefills completed, so we need to be careful
        while self._preempted_requests:
            if len(requests) == self._max_micro_batch_size:
                break

            request = self._preempted_requests.pop(0)

            if not self._is_prefill_complete(request):
                running_prefills.append(request)
                continue

            next_num_q_tokens = self._get_request_next_num_q_tokens(
                request, total_num_q_tokens
            )
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens

            if next_num_q_tokens == 0:
                skipped_requests.append(request)
                continue

            while not self._can_allocate_request(request):
                if self._preempted_requests:
                    victim_request = self._preempted_requests.pop(-1)
                    victim_request.restart()
                    self.free(victim_request.id)
                    self._request_queue = [victim_request] + self._request_queue
                else:
                    request.restart()
                    self.free(request.id)
                    self._request_queue = [request] + self._request_queue
                    break
            else:
                self._allocate_request(request)
                assert request.is_prefill_complete
                total_num_q_tokens += next_num_q_tokens
                requests.append(request)
                num_q_tokens.append(next_num_q_tokens)
                num_kv_tokens.append(next_num_kv_tokens)

        for request in running_prefills:
            assert not request.is_prefill_complete

            next_num_q_tokens = self._get_request_next_num_q_tokens(
                request, total_num_q_tokens
            )
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens

            if next_num_q_tokens == 0:
                skipped_requests.append(request)
                continue

            total_num_q_tokens += next_num_q_tokens
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)
            num_kv_tokens.append(next_num_kv_tokens)

        # re-add the skipped requests, but make sure that we add them to the
        # front of the queue so that they are scheduled first and we maintain FIFO ordering
        self._preempted_requests = skipped_requests + self._preempted_requests
        self._preempted_requests = sorted(
            self._preempted_requests, key=lambda req: req.arrived_at
        )

        while self._request_queue:
            if len(self._allocation_map) == self._config.batch_size_cap:
                break

            if len(requests) == self._max_micro_batch_size:
                break

            request = self._request_queue[0]

            if not self._can_allocate_request(request):
                break

            next_num_q_tokens = self._get_request_next_num_q_tokens(
                request, total_num_q_tokens
            )
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens

            if next_num_q_tokens == 0:
                break

            self._request_queue.pop(0)

            self._allocate_request(request)

            # all new requests will have a prefill
            total_num_q_tokens += next_num_q_tokens
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)
            num_kv_tokens.append(next_num_kv_tokens)

        if not requests:
            return

        return Batch(self._replica_id, requests, num_q_tokens, num_kv_tokens)



from dataclasses import dataclass, field

from vidur.config.base_fixed_config import BaseFixedConfig
from vidur.logger import init_logger
from vidur.types import DeviceSKUType, NodeSKUType

logger = init_logger(__name__)


@dataclass
class BaseNodeSKUConfig(BaseFixedConfig):
    num_devices_per_node: int


@dataclass
class A40PairwiseNvlinkNodeSKUConfig(BaseNodeSKUConfig):
    device_sku_type: DeviceSKUType = DeviceSKUType.A40
    num_devices_per_node: int = 8

    @staticmethod
    def get_type():
        return NodeSKUType.A40_PAIRWISE_NVLINK


@dataclass
class A100PairwiseNvlinkNodeSKUConfig(BaseNodeSKUConfig):
    device_sku_type: DeviceSKUType = DeviceSKUType.A100
    num_devices_per_node: int = 4

    @staticmethod
    def get_type():
        return NodeSKUType.A100_PAIRWISE_NVLINK


@dataclass
class H100PairwiseNvlinkNodeSKUConfig(BaseNodeSKUConfig):
    device_sku_type: DeviceSKUType = DeviceSKUType.H100
    num_devices_per_node: int = 4

    @staticmethod
    def get_type():
        return NodeSKUType.H100_PAIRWISE_NVLINK


@dataclass
class A100DgxNodeSKUConfig(BaseNodeSKUConfig):
    device_sku_type: DeviceSKUType = DeviceSKUType.A100
    num_devices_per_node: int = 8

    @staticmethod
    def get_type():
        return NodeSKUType.A100_DGX


@dataclass
class H100DgxNodeSKUConfig(BaseNodeSKUConfig):
    device_sku_type: DeviceSKUType = DeviceSKUType.H100
    num_devices_per_node: int = 8

    @staticmethod
    def get_type():
        return NodeSKUType.H100_DGX



class AttentionInput:
    def __init__(
        self,
        prefill_chunk_size: int,
        kv_cache_size: int,
        batch_size: int,
        is_prefill: bool,
    ):
        self.prefill_chunk_size = prefill_chunk_size
        self.kv_cache_size = kv_cache_size
        self.batch_size = batch_size
        self.is_prefill = is_prefill

    def is_valid(self, max_seq_len: int):
        if self.is_prefill:
            if self.batch_size != 1:
                return False
            elif self.prefill_chunk_size == 0:
                return False
            elif self.prefill_chunk_size + self.kv_cache_size > max_seq_len:
                return False
        else:
            if self.prefill_chunk_size > 0:
                return False
            elif self.kv_cache_size == 0:
                return False
            elif self.kv_cache_size > max_seq_len:
                return False
        return True

    def is_under_memory_limit(self, max_num_tokens: int):
        return (
            self.batch_size * (self.kv_cache_size + self.prefill_chunk_size)
            <= max_num_tokens
        )

    def __str__(self):
        return f"prefill_chunk_size: {self.prefill_chunk_size}, kv_cache_size: {self.kv_cache_size}, batch_size: {self.batch_size}, is_prefill: {self.is_prefill}"



from typing import Tuple

from vidur.entities.base_entity import BaseEntity
from vidur.logger import init_logger

logger = init_logger(__name__)


# a decorator which checks if the request has been scheduled
def check_scheduled(func):
    def wrapper(self, *args, **kwargs):
        if not self._scheduled:
            raise ValueError("Request has not been scheduled yet")
        return func(self, *args, **kwargs)

    return wrapper


def check_completed(func):
    def wrapper(self, *args, **kwargs):
        if not self._completed:
            raise ValueError("Request has not been completed yet")
        return func(self, *args, **kwargs)

    return wrapper


class Request(BaseEntity):
    def __init__(
        self,
        arrived_at: float,
        num_prefill_tokens: int,
        num_decode_tokens: int,
    ):
        self._id = Request.generate_id()
        self._arrived_at = arrived_at
        self._num_prefill_tokens = num_prefill_tokens
        self._num_decode_tokens = num_decode_tokens

        self._num_processed_tokens = 0
        self._num_stage_processed_tokens = 0

        self._scheduled_at = 0
        self._execution_time = 0
        self._model_execution_time = 0
        self._scheduling_delay = 0
        self._preempted_time = 0
        self._completed_at = 0
        self._prefill_completed_at = 0
        self._latest_stage_scheduled_at = 0
        self._latest_stage_completed_at = 0
        self._latest_iteration_scheduled_at = 0
        self._latest_iteration_completed_at = 0
        self._latest_iteration_scheduling_delay = 0
        self._num_processed_iterations = 0

        self._scheduled = False
        self._preempted = False
        self._completed = False
        self._is_prefill_complete = False
        self._is_prefill_stage_complete = False

        self._num_restarts = 0

    @property
    def size(self) -> Tuple[int, int]:
        return (self._num_prefill_tokens, self._num_decode_tokens)

    @property
    @check_scheduled
    def scheduled_at(self) -> float:
        return self._scheduled_at

    @property
    @check_scheduled
    def latest_stage_scheduled_at(self) -> float:
        return self._latest_stage_scheduled_at

    @property
    @check_scheduled
    def latest_stage_completed_at(self) -> float:
        return self._latest_stage_completed_at

    @property
    @check_scheduled
    def latest_iteration_scheduled_at(self) -> float:
        return self._latest_iteration_scheduled_at

    @property
    @check_scheduled
    def latest_iteration_completed_at(self) -> float:
        return self._latest_iteration_completed_at

    @property
    def num_processed_iterations(self) -> int:
        return self._num_processed_iterations

    @property
    @check_scheduled
    def latest_iteration_scheduling_delay(self) -> float:
        return self._latest_iteration_scheduling_delay

    @property
    @check_scheduled
    def prefill_completed_at(self) -> float:
        return self._prefill_completed_at

    @property
    @check_scheduled
    def scheduling_delay(self) -> float:
        return self._scheduling_delay

    @property
    @check_scheduled
    def preempted_time(self) -> float:
        return self._preempted_time

    @property
    @check_completed
    def completed_at(self) -> float:
        return self._completed_at

    @property
    @check_scheduled
    def e2e_time(self) -> float:
        return self._completed_at - self._arrived_at

    @property
    @check_scheduled
    def e2e_time_normalized(self) -> float:
        return self.e2e_time / self.num_decode_tokens

    @property
    @check_scheduled
    def execution_time(self) -> float:
        return self._execution_time

    @property
    @check_scheduled
    def execution_time_normalized(self) -> float:
        return self._execution_time / self.num_decode_tokens

    @property
    @check_scheduled
    def model_execution_time(self) -> float:
        return self._model_execution_time

    @property
    @check_scheduled
    def model_execution_time_normalized(self) -> float:
        return self._model_execution_time / self.num_decode_tokens

    @property
    def arrived_at(self) -> float:
        return self._arrived_at

    @property
    def num_prefill_tokens(self) -> int:
        return self._num_prefill_tokens

    @property
    def num_decode_tokens(self) -> int:
        return self._num_decode_tokens

    @property
    def pd_ratio(self) -> float:
        return self._num_prefill_tokens / self._num_decode_tokens

    @property
    def num_processed_tokens(self) -> int:
        return self._num_processed_tokens

    @property
    def num_stage_processed_tokens(self) -> int:
        return self._num_stage_processed_tokens

    @property
    def total_tokens(self) -> int:
        return self._num_prefill_tokens + self._num_decode_tokens

    @property
    def num_processed_prefill_tokens(self) -> int:
        return min(self._num_processed_tokens, self._num_prefill_tokens)

    @property
    def num_processed_decode_tokens(self) -> int:
        return max(self._num_processed_tokens - self._num_prefill_tokens, 0)

    @property
    def scheduled(self) -> bool:
        return self._scheduled

    @property
    def preempted(self) -> bool:
        return self._preempted and not self._completed

    @property
    def completed(self) -> bool:
        return self._completed

    @property
    def num_restarts(self) -> int:
        return self._num_restarts

    @property
    def is_prefill_complete(self) -> bool:
        return self._is_prefill_complete

    @property
    def is_prefill_stage_complete(self) -> bool:
        return self._is_prefill_stage_complete

    @property
    def has_started_decode(self) -> bool:
        return self._num_processed_tokens > self._num_prefill_tokens + 1

    def on_batch_schedule(
        self,
        time: float,
    ) -> None:
        self._latest_iteration_scheduled_at = time
        self._latest_iteration_scheduling_delay = (
            time - self._latest_iteration_completed_at
        )

        if self._scheduled:
            return

        if self._num_restarts > 0:
            self._scheduled = True
            return

        self._scheduled_at = time
        self._scheduling_delay = time - self._arrived_at
        self._scheduled = True

    def on_batch_end(
        self,
        time: float,
        num_tokens_processed: int,
    ) -> None:
        assert num_tokens_processed > 0
        assert not self._completed

        self._num_processed_tokens += num_tokens_processed
        assert self._num_processed_tokens <= self.total_tokens

        if self._num_processed_tokens == self._num_prefill_tokens:
            assert (
                self._is_prefill_stage_complete
            ), f"Prefill num tokens stage processed: {self._num_stage_processed_tokens}, num processed tokens: {self._num_processed_tokens}"

            self._is_prefill_complete = True
            # we get one decode token when the prefill processing completes
            self._num_processed_tokens += 1

            # we must record the prefill completion time only in the first time
            # in the subsequent restarts, we keep adding the previously decoded
            # tokens to the prefill tokens - that is irrelevant to the original prefill
            if self._prefill_completed_at == 0:
                self._prefill_completed_at = time

        # check if request is completed
        if self._num_processed_tokens == self.total_tokens:
            self._completed_at = time
            self._completed = True
            logger.debug(f"Request {self._id} completed at {self._completed_at}")

        self._latest_iteration_completed_at = time
        self._num_processed_iterations += 1

    def on_batch_stage_schedule(
        self,
        time: float,
    ) -> None:
        self._latest_stage_scheduled_at = time
        if self._latest_stage_completed_at == 0:
            self._preempted_time = 0
        else:
            self._preempted_time += time - self._latest_stage_completed_at
        self._preempted = False

    def on_batch_stage_end(
        self,
        time: float,
        stage_id: int,
        execution_time: float,
        model_execution_time: float,
        num_tokens_stage_processed: int,
    ) -> None:
        self._execution_time += execution_time
        self._model_execution_time += model_execution_time
        self._latest_stage_completed_at = time
        self._preempted = True

        if not stage_id == 0:
            return

        self._num_stage_processed_tokens += num_tokens_stage_processed
        assert self._num_stage_processed_tokens <= self.total_tokens
        if self._num_stage_processed_tokens == self._num_prefill_tokens:
            self._is_prefill_stage_complete = True
            self._num_stage_processed_tokens += 1

    def to_dict(self) -> dict:
        return {
            "id": self._id,
            "arrived_at": self._arrived_at,
            "execution_time": self._execution_time,
            "model_execution_time": self._model_execution_time,
            "scheduled_at": self._scheduled_at,
            "scheduling_delay": self._scheduling_delay,
            "preempted_time": self._preempted_time,
            "completed_at": self._completed_at,
            "num_prefill_tokens": self._num_prefill_tokens,
            "num_decode_tokens": self._num_decode_tokens,
            "num_processed_tokens": self._num_processed_tokens,
            "scheduled": self._scheduled,
            "preempted": self._preempted,
            "completed": self._completed,
            "latest_stage_scheduled_at": self._latest_stage_scheduled_at,
            "latest_stage_completed_at": self._latest_stage_completed_at,
            "latest_iteration_scheduled_at": self._latest_iteration_scheduled_at,
            "latest_iteration_completed_at": self._latest_iteration_completed_at,
            "num_restarts": self._num_restarts,
        }

    def restart(self):
        logger.debug(f"Restarting request {self._id}")

        # when we restart the request, we can process all the previously
        # decoded tokens in parallel (i.e., we can prefill all the tokens)
        total_tokens = self._num_prefill_tokens + self._num_decode_tokens
        self._num_prefill_tokens = self._num_processed_tokens
        self._num_decode_tokens = total_tokens - self._num_prefill_tokens

        self._num_processed_tokens = 0
        self._num_processed_iterations = 0
        self._scheduled = False
        self._preempted = False
        self._completed = False
        self._is_prefill_complete = False

        self._num_restarts += 1



from scipy.stats import gamma

from vidur.config import GammaRequestIntervalGeneratorConfig
from vidur.request_generator.base_request_interval_generator import (
    BaseRequestIntervalGenerator,
)


class GammaRequestIntervalGenerator(BaseRequestIntervalGenerator):

    def __init__(self, config: GammaRequestIntervalGeneratorConfig):
        super().__init__(config)

        cv = self.config.cv
        self.qps = self.config.qps
        self.gamma_shape = 1.0 / (cv**2)

    def get_next_inter_request_time(self) -> float:
        gamma_scale = 1.0 / (self.qps * self.gamma_shape)
        return gamma.rvs(self.gamma_shape, scale=gamma_scale)



try:
    from ._version import __version__, __version_tuple__
except Exception as e:
    import warnings

    warnings.warn(f"Failed to read commit hash:\n{e}", RuntimeWarning, stacklevel=2)

    __version__ = "dev"
    __version_tuple__ = (0, 0, __version__)



import gc
import os
from typing import Optional

import ray
import torch

from vidur.logger import init_logger
from vidur.profiling.collectives.collectives_input import CollectivesInput
from vidur.profiling.collectives.collectives_wrapper import CollectiveWrapper

logger = init_logger(__name__)


@ray.remote(num_gpus=1)
class BenchmarkRunner:
    def __init__(self, gpu_id: int, max_gpus_per_node: int, head_ip: str) -> None:
        self._gpu_id = gpu_id
        self._max_devices_per_node = max_gpus_per_node
        self._set_cuda_visible_devices()
        self._last_num_workers_per_node = None
        self._last_num_workers = None
        self._head_ip = head_ip

    def _set_cuda_visible_devices(self) -> None:
        os.environ["CUDA_VISIBLE_DEVICES"] = str(
            self._gpu_id % self._max_devices_per_node
        )
        # set additional nccl env vars
        # This env var set by Ray causes exceptions with graph building.
        os.environ.pop("NCCL_ASYNC_ERROR_HANDLING", None)
        # Required for properly capturing nccl ops
        os.environ["NCCL_GRAPH_MIXING_SUPPORT"] = "0"
        os.environ["KINETO_LOG_LEVEL"] = "5"
        os.environ["NCCL_IGNORE_DISABLED_P2P"] = "1"

    def run_collective(
        self,
        collectives_input: CollectivesInput,
    ) -> Optional[dict]:
        if (
            collectives_input.num_workers != self._last_num_workers
            or collectives_input.num_workers_per_node != self._last_num_workers_per_node
        ) and torch.distributed.is_initialized():
            torch.distributed.destroy_process_group()

        rank = self._get_rank(
            collectives_input.num_workers, collectives_input.num_workers_per_node
        )
        if rank is None:
            return None

        if (
            collectives_input.num_workers != self._last_num_workers
            or collectives_input.num_workers_per_node != self._last_num_workers_per_node
        ):
            self._init_communication(
                collectives_input.comm_id,
                rank,
                collectives_input.num_workers,
                collectives_input.num_workers_per_node,
            )
            self._last_num_workers = collectives_input.num_workers
            self._last_num_workers_per_node = collectives_input.num_workers_per_node

        wrapper = CollectiveWrapper(
            rank,
            collectives_input.num_workers,
            collectives_input.comm_id,
            collectives_input.collective_size,
            collectives_input.collective,
            collectives_input.num_workers_per_node,
            self._max_devices_per_node,
        )
        stats = wrapper.profile()
        del wrapper
        gc.collect()
        return stats

    def _init_communication(
        self, comm_id: int, rank: int, num_workers: int, devices_per_node: int
    ):
        logger.info(
            f"Initializing gpu id: {self._gpu_id}, Rank: {rank}, num_workers: {num_workers}, comm_id: {comm_id}, "
            f"devices_per_node: {devices_per_node}, max_devices_per_node: {self._max_devices_per_node}, "
            f"ip_addr: {ray.util.get_node_ip_address()}, CUDA_VISIBLE_DEVICES: {os.environ['CUDA_VISIBLE_DEVICES']}"
        )

        torch.distributed.init_process_group(
            backend="nccl",
            rank=rank,
            world_size=num_workers,
            init_method=f"tcp://{self._head_ip}:{comm_id}",
        )

    def _get_rank(self, num_workers: int, devices_per_node: int):
        assert self._max_devices_per_node >= devices_per_node
        assert self._max_devices_per_node % devices_per_node == 0
        assert num_workers % devices_per_node == 0 or num_workers < devices_per_node

        num_nodes = num_workers // devices_per_node
        current_node = self._gpu_id // self._max_devices_per_node

        if current_node >= num_nodes:
            return None

        local_gpu_id = self._gpu_id % self._max_devices_per_node

        # # scatter devices uniformly across the node
        # node_devices = list(range(self._max_devices_per_node))
        # device_offset = self._max_devices_per_node // devices_per_node

        # # selected devices for this worker
        # selected_devices = node_devices[::device_offset]

        # pack devices in order
        selected_devices = list(range(devices_per_node))

        if local_gpu_id not in selected_devices:
            return None

        # rank of this worker
        rank = current_node * devices_per_node + selected_devices.index(local_gpu_id)

        return rank



from vidur.entities.batch import Batch
from vidur.scheduler.replica_scheduler.base_replica_scheduler import (
    BaseReplicaScheduler,
)


class FasterTransformerReplicaScheduler(BaseReplicaScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._pending_free_map = {}

    def on_batch_end(self, batch: Batch) -> None:
        super().on_batch_end(batch)

        if batch.have_all_requests_completed:
            # free memory for all requests at once
            self.free_batch(batch)
            self.free(*self._pending_free_map.pop(batch.id, []))
        else:
            self._preempted_batches.append(batch)

    def _generate_next_batch_from_preempted(self, preempted_batch: Batch) -> Batch:
        requests = []
        num_q_tokens = []
        num_kv_tokens = []

        for request in preempted_batch.requests:
            if request.completed:
                continue
            next_num_q_tokens = self._get_request_next_num_q_tokens(request)
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)
            num_kv_tokens.append(next_num_kv_tokens)

        if not requests:
            return

        return Batch(self._replica_id, requests, num_q_tokens, num_kv_tokens)

    def _get_next_batch(self, time: float) -> Batch:
        if self._preempted_batches:
            preempted_batch = self._preempted_batches.pop(0)
            return self._generate_next_batch_from_preempted(preempted_batch)

        requests = []
        num_q_tokens = []
        num_kv_tokens = []

        while self._request_queue:
            if len(requests) == self._max_batch_size:
                break

            if not self.can_allocate(self._max_blocks_per_sequence):
                break

            request = self._request_queue.pop(0)
            self.allocate(request.id, self._max_blocks_per_sequence)
            next_num_q_tokens = self._get_request_next_num_q_tokens(request)
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)
            num_kv_tokens.append(next_num_kv_tokens)

        if not requests:
            return

        return Batch(self._replica_id, requests, num_q_tokens, num_kv_tokens)



import argparse
import datetime
import os

import pandas as pd
import ray
from tqdm import tqdm

from vidur.logger import init_logger
from vidur.profiling.collectives.benchmark_runner import BenchmarkRunner
from vidur.profiling.utils import get_collectives_inputs

logger = init_logger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(description="MLP Profiling")
    parser.add_argument(
        "--num_workers_per_node_combinations",
        type=int,
        nargs="+",
        default=[8],
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="profiling_outputs",
        help="Output directory for profiling results",
    )
    parser.add_argument(
        "--max_collective_size",
        type=int,
        default=4096 * 8192,
        help="Maximum number of elements involved in the collective",
    )
    parser.add_argument(
        "--collective",
        default="all_reduce",
        choices=["all_reduce", "send_recv"],
        help="Collective to profile",
    )
    args = parser.parse_args()

    args.output_dir = f"{args.output_dir}/collective/{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
    os.makedirs(args.output_dir, exist_ok=True)

    return args


def create_runner_pool():
    total_gpus_available = int(ray.cluster_resources()["GPU"])
    logger.info(f"Total GPUs available: {total_gpus_available}")

    assert total_gpus_available > 0, "No GPUs available"

    all_node_ips = [x["NodeName"] for x in ray.nodes()]
    logger.info(f"All node IPs: {all_node_ips}")

    assert len(all_node_ips) > 0, "No nodes available"

    num_nodes = len(all_node_ips)
    gpus_per_node = total_gpus_available // len(all_node_ips)

    runner_pool = []
    for gpu_id in range(total_gpus_available):
        node_ip = all_node_ips[gpu_id // gpus_per_node]
        runner_pool.append(
            BenchmarkRunner.options(
                resources={
                    f"node:{node_ip}": 0.01,
                }
            ).remote(gpu_id, gpus_per_node, all_node_ips[0])
        )
    return total_gpus_available, num_nodes, runner_pool


def main():
    args = parse_args()

    ray.init()

    total_gpus_available, num_nodes, runner_pool = create_runner_pool()

    all_results = []

    collectives_inputs = get_collectives_inputs(
        num_nodes,
        args.num_workers_per_node_combinations,
        args.max_collective_size,
        args.collective,
        total_gpus_available,
    )

    for collectives_input in tqdm(collectives_inputs):
        promises = []
        for gpu_id in range(total_gpus_available):
            promise = runner_pool[gpu_id].run_collective.remote(collectives_input)
            promises.append(promise)

        for gpu_id in range(int(total_gpus_available)):
            result = ray.get([promises[gpu_id]])[0]
            if result and gpu_id == 0:
                all_results.append(result)

        ray.get(promises)

    # filter none results
    all_results = [x for x in all_results if x is not None]

    df = pd.DataFrame(all_results)
    # the time_stats column is a dict, so we need to expand it into columns recursively and add prefix

    df = (
        pd.json_normalize(df["time_stats"])
        .add_prefix("time_stats.")
        .join(df.drop(columns=["time_stats"]))
    )

    # write results to a csv file
    df.to_csv(f"{args.output_dir}/{args.collective}.csv", compression="xz")


if __name__ == "__main__":
    main()



import argparse
import datetime
import itertools
import os
from typing import Any, List

import pandas as pd
import ray
import yaml
from tqdm import tqdm

from vidur.profiling.common.model_config import ModelConfig
from vidur.profiling.mlp.mlp_wrapper import MlpWrapper
from vidur.profiling.utils import ProfileMethod, get_num_tokens_to_profile


def parse_args():
    parser = argparse.ArgumentParser(description="MLP Profiling")
    parser.add_argument(
        "--disable_ray",
        action="store_true",
        help="Disable Ray",
    )
    parser.add_argument(
        "--num_gpus",
        type=int,
        default=8,
        help="Number of GPUs to use for profiling",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="profiling_outputs",
        help="Output directory for profiling results",
    )
    parser.add_argument(
        "--models",
        type=str,
        nargs="+",
        default=[
            "meta-llama/Llama-3-8B",
            "meta-llama/Llama-3-70B",
        ],
        help="Models to profile",
    )
    parser.add_argument(
        "--num_tensor_parallel_workers",
        type=int,
        nargs="+",
        default=[1, 2, 4, 8],
        help="Number of tensor parallel workers to profile",
    )
    parser.add_argument(
        "--max_tokens",
        type=int,
        default=4096,
        help="Maximum number of tokens to profile",
    )
    parser.add_argument(
        "--profile_method",
        default="record_function",
        choices=[e.value for e in ProfileMethod],
        help="Method to use for measuring time taken by operations (default: %(default)s)",
    )
    args = parser.parse_args()

    args.output_dir = (
        f"{args.output_dir}/mlp/{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
    )
    os.makedirs(args.output_dir, exist_ok=True)

    return args


def profile_model(
    args: argparse.Namespace, model: str, num_tokens_to_profile: List[int], pbar: Any
):
    model_config = ModelConfig.from_model_name(model)

    promises = []
    all_results = []

    model_wrapper_actor = ray.remote(
        num_cpus=1,
        num_gpus=1,
    )(
        MlpWrapper,
    ).options(runtime_env={"env_vars": {"KINETO_LOG_LEVEL": "5"}})

    for num_tensor_parallel_workers in args.num_tensor_parallel_workers:
        if model_config.no_tensor_parallel and num_tensor_parallel_workers > 1:
            pbar.update(len(num_tokens_to_profile))
            continue

        model_wrappers = [
            model_wrapper_actor.remote(
                model_config,
                num_tensor_parallel_workers,
                args.profile_method,
                rank,
                args.output_dir,
            )
            for rank in range(args.num_gpus)
        ]
        for num_tokens in num_tokens_to_profile:
            worker_id = len(promises)
            promise = model_wrappers[worker_id].profile.remote(
                num_tokens,
            )
            promises.append(promise)

            if len(promises) >= args.num_gpus:
                results = ray.get(promises)
                all_results.extend(results)
                promises = []

            pbar.update(1)

    results = ray.get(promises)
    all_results.extend(results)

    df = pd.DataFrame(all_results)
    # the time_stats column is a dict, so we need to expand it into columns recursively and add prefix
    df = (
        pd.json_normalize(df["time_stats"])
        .add_prefix("time_stats.")
        .join(df.drop(columns=["time_stats"]))
    )

    return df


def main():
    args = parse_args()
    yaml.dump(vars(args), open(f"{args.output_dir}/config.yaml", "w"))

    num_tokens_to_profile = get_num_tokens_to_profile(args.max_tokens)

    total_combos = itertools.product(
        args.models,
        num_tokens_to_profile,
        args.num_tensor_parallel_workers,
    )

    pbar = tqdm(total=len(list(total_combos)))

    for model in args.models:
        result_df = profile_model(
            args,
            model,
            num_tokens_to_profile,
            pbar,
        )
        # model name would contain '/', so create a directory as required
        os.makedirs(f"{args.output_dir}/{model}", exist_ok=True)
        result_df.to_csv(
            f"{args.output_dir}/{model}/mlp.csv", index=False, compression="xz"
        )


if __name__ == "__main__":
    main()
